# -*- coding:utf-8 -*-
"""
This module contains functions that can be called on the server side to create some side effects
Those functions are aimed to be called by the EntityIterator 'call' method
Each function MUST g and ctx as 2 first arguments
g is the graph
ctx is the "context" that is filled when EntityIterator 'as' method is used

From the client side it allows to build queries such as :
g.V().aka('x').outV(label='has_parent').outV(label='has_parent').aka('y').call('add_edge', 'x', 'y', label='has_grandparent')

The final call will :
- effectively be run by the server
- potentially have side effects on the graph

As you can see, in the call method, there's no need to pass '_g' and '_ctxs' as those 2 variables are hold by the EntityIterator
The 'call' method will be in charge of looking for the right function and calling function passing g/ctx/ and other parameters
"""


def add_edge(_g, _ctxs, _ref1, _ref2, **data):
    edge_defns = []
    for _ctx in _ctxs:
        node1 = _ctx[_ref1]
        node2 = _ctx[_ref2]
        edge_defns.append((node1, node2, data))
    _g.bulk_add_edge(edge_defns, silent=True)



# -*- coding:utf-8 -*-

from grapheekdb.lib.exceptions import GrapheekInvalidDataTypeException
from grapheekdb.lib.expressions import VALID_ATTRIBUTE_REGEX

try:  # pragma : no cover
    LONG = long
except NameError:  # pragma : no cover
    LONG = int
NONETYPE = type(None)
try:  # pragma : no cover
    UNICODE = unicode
except NameError:  # pragma : no cover
    UNICODE = str


def check_valid_value(value):
    if isinstance(value, (bool, float, int, str, LONG, NONETYPE, UNICODE)):
        return True
    elif isinstance(value, list):
        return all([check_valid_value(subvalue) for subvalue in value])
    elif isinstance(value, dict):
        return check_valid_data(value, check_dict=False)
    return False


def check_valid_data(data, check_dict=True):
    # Checking that data is a dictionnary
    if check_dict and not isinstance(data, dict):
        raise GrapheekInvalidDataTypeException('data must be a dictionnary')
    # Ckecking that data keys match '^[a-zA-A][a-zA-Z0-9]*$' regular expression
    if not(all(VALID_ATTRIBUTE_REGEX.match(key) for key in data)):
        raise GrapheekInvalidDataTypeException('a data key must only contain letter and digit and start with a letter')
    # Ckecking that data keys don't contain a double underscore (could mess lookup)
    if any('__' in key for key in data):
        raise GrapheekInvalidDataTypeException('a data key must not contain a double underscore')
    # Ckecking that data keys don't end with an underscore (could mess lookup)
    if any(key.endswith('_') for key in data):
        raise GrapheekInvalidDataTypeException('a data key must not end with an underscore')
    # Values validations  :
    if not all([check_valid_value(value) for value in list(data.values())]):
        raise GrapheekInvalidDataTypeException('value must be one of following type : boolean/float/integer/long/none/string/unicode/dict/list')
    return True



# -*- coding:utf-8 -*-

import sys

import msgpack

from kyotocabinet import DB

from grapheekdb.backends.data.base import BaseGraph
from grapheekdb.lib.exceptions import GrapheekDataException

from grapheekdb.lib.undef import UNDEFINED

PYTHON2 = sys.version_info.major == 2

class GrapheekDataKyotoCabinetException(GrapheekDataException):
    pass


class GrapheekDataKyotoCabinetInitFailureException(GrapheekDataKyotoCabinetException):
    pass


class KyotoCabinetGraph(BaseGraph):

    def __init__(self, path):
        # create the database object
        self._path = path
        self._db = DB()
        # open the database
        if not self._db.open(path, DB.OREADER | DB.OWRITER | DB.OCREATE):
            raise GrapheekDataKyotoCabinetInitFailureException(str(self._db.error()))
        super(KyotoCabinetGraph, self).__init__()
        self._ensure_prepared()
        self._closed = False

    # Start method overriding :

    def _db_close(self):
        if not self._closed:
            self._db.close()

    def _transaction_begin(self):
        self._db.begin_transaction()
        return True

    def _transaction_commit(self, txn):
        self._db.end_transaction(True)

    def _transaction_rollback(self, txn):
        self._db.end_transaction(False)

    def _has_key(self, key):
        return self._db.check(key) >= 0

    def _get(self, txn, key):
        raw_data = self._db.get(key)
        if raw_data is None:
            return UNDEFINED  # Not returning None, as None is a valid value
        return msgpack.loads(raw_data, encoding='utf8')

    def _bulk_get(self, txn, keys):
        result = {}
        key_raw_datas = self._db.get_bulk(keys)
        for key, raw_data in list(key_raw_datas.items()):
            if PYTHON2:  # pragma : no cover
                k = key
            else:  # pragma : no cover
                k = str(key, encoding='utf8')
            result[k] = msgpack.loads(raw_data, encoding='utf8')
        return result

    def _set(self, txn, key, value):
        res = self._db.set(key, msgpack.dumps(value, encoding='utf8'))
        if not(res):  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _bulk_set(self, txn, updates):
        dic = {}
        for key, value in list(updates.items()):
            dic[key] = msgpack.dumps(value, encoding='utf8')
        res = self._db.set_bulk(dic)
        if res == -1:  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _remove(self, txn, key):
        # Contrary to LocalMemoryGraph implementation, it is not needed to wrap
        # key removal in try.. except because KyotoCabinet only send "False"
        # when key does not exist
        # Thus ... _removemethod is idempotent (cf LocalMemoryGraph _remove method comment)
        self._db.remove(key)

    def _bulk_remove(self, txn, keys):
        res = self._db.remove_bulk(list(keys))
        if res == -1:  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _remove_prefix(self, txn, prefix):
        keys = self._db.match_prefix(prefix)
        self._db.remove_bulk(keys)

    # overriding list methods
    # looks like a bucket of hacks, and yes indeed it is :)
    # btw, it REALLY improves performance if we compare to default implementation which,
    # in the case of KyotoCabinet would involve msgpack deserialization followed by a serialization

    def _init_lst(self, txn, key):
        res = self._db.set(key, '')
        if not(res):  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _get_lst(self, txn, key):
        value = self._db.get(key)
        if value is None:
            return UNDEFINED
        # look _append_to_lst code below to understand why a split is done
        # And why resulting list is sliced from 1
        if PYTHON2:  # pragma : no cover
            return list(map(int, value.split('|')[1:]))
        return list(map(int, str(value, encoding='utf8').split('|')[1:]))  # pragma : no cover

    def _set_lst(self, txn, key, values):
        newval = '|'.join([str(value) for value in values])
        res = self._db.set(key, '|' + newval)
        if not(res):  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _bulk_get_lst(self, txn, keys):
        dic_values = self._db.get_bulk(keys)
        results = []
        for key in keys:
            if PYTHON2:  # pragma : no cover
                values = dic_values.get(key, UNDEFINED)
            else:  # pragma : no cover
                values = dic_values.get(bytes(key, encoding='utf8'), UNDEFINED)
            if values == UNDEFINED:
                results.append([])
            else:
                if PYTHON2:  # pragma : no cover
                    results.append(list(map(int, values.split('|')[1:])))
                else:  # pragma : no cover
                    results.append(list(map(int, str(values, encoding='utf8').split('|')[1:])))
        return results

    def _append_to_lst(self, txn, key, value):
        self._db.append(key, '|' + str(value))

    def _bulk_append_to_lst(self, txn, key, values):
        newval = '|'.join([str(value) for value in values])
        self._db.append(key, '|' + newval)

    def _remove_from_lst(self, txn, key, value):
        old = self._db.get(key)
        if not PYTHON2:  # pragma : no cover
            old = str(old, encoding='utf8')
        # Caution : we are only removing ONE occurence
        # This is voluntary
        # For instance, it lst contains neighbour node, we need to remove only one occurence
        # cause current entity and neighbour node can be linked multiple time
        new = old.replace('|%s' % value, '', 1)
        if new == old:
            raise ValueError("list.remove(x): x not in list")
        res = self._db.set(key, new)
        if not(res):  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res

    def _bulk_remove_from_lst(self, txn, key, values):
        assert(len(values))
        old = self._db.get(key)
        if PYTHON2:  # pragma : no cover
            new = old
        else:  # pragma : no cover
            new = str(old, encoding='utf8')
        for value in values:
            new = new.replace('|%s' % value, '', 1)
        if new == old:  # pragma : no cover
            raise ValueError("list.remove(x): x not in list")
        res = self._db.set(key, new)
        if not(res):  # pragma : no cover
            raise GrapheekDataKyotoCabinetException('KyotoCabinet : error while saving')
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import traceback
import msgpack
import argparse
from collections import OrderedDict

try:  # pragma : no cover
    from ConfigParser import RawConfigParser  # Python 2
except:  # pragma : no cover
    from configparser import RawConfigParser  # Python 3

try:  # pragma : no cover
    LONG = long
except NameError:  # pragma : no cover
    LONG = int
NONETYPE = type(None)
try:  # pragma : no cover
    UNICODE = unicode
except NameError:  # pragma : no cover
    UNICODE = str

import zmq

from grapheekdb.backends.data.base import EntityAggregate, EntityIterator, Node, Edge
from grapheekdb import __version__
from grapheekdb.lib.exceptions import GrapheekMarshallingException


def marshall(item):
    if isinstance(item, (bool, float, int, str, LONG, NONETYPE, UNICODE)):
        return item
    if isinstance(item, bytes):  # pragma : no cover
        # this should only occur in Python3
        return str(item, encoding='utf8')
    elif isinstance(item, (list, tuple)):
        return [marshall(x) for x in item]
    elif isinstance(item, (dict, OrderedDict)):
        items = [(marshall(key), marshall(value)) for key, value in list(item.items())]
        # Not returning : dict(items)
        # Because marshall(key) could be an unhashable type (such as dict)
        # contrary to key...
        return {
            '__': 'd',
            'd': items
        }
    elif isinstance(item, Node):
        return {
            '__': 'n',
            '_i': item.get_id(),
            'd': marshall(item.data())
        }
    elif isinstance(item, Edge):
        return {
            '__': 'e',
            '_i': item.get_id(),
            'd': marshall(item.data())
        }
    elif isinstance(item, (EntityAggregate, EntityIterator)):
        return [marshall(x) for x in item]
    raise GrapheekMarshallingException('Unknown type or instance : %s - %s' % (type(item), item))

def fixbytes(item):
    if isinstance(item, (bool, float, int, str, LONG, NONETYPE, UNICODE)):
        return item
    if isinstance(item, bytes):  # pragma : no cover
        # this should only occur in Python3
        return str(item, encoding='utf8')
    elif isinstance(item, (list, tuple)):
        return [fixbytes(x) for x in item]
    elif isinstance(item, dict):
        return dict([(fixbytes(key), fixbytes(value)) for key, value in list(item.items())])
    raise GrapheekMarshallingException('Unknown type or instance : %s - %s' % (type(item), item))  # pragma : no cover


def runserver(_address, _backend, _scripts, _context, **params):
    class_module_name, class_name = _backend.rsplit('.', 1)
    # START -- Following lines may raise an exception if data module cannot be imported
    # I'm letting exception propagate so that the user can fix path (or other potential errors)
    __import__(class_module_name)
    class_module = sys.modules[class_module_name]
    # -- END
    GraphClass = getattr(class_module, class_name)
    g = GraphClass(**params)
    try:
        # Setting script module paths :
        g.setup_server_scripts(*_scripts.split(':'))
        # --
        if _context is None:  # pragma : no cover
            _context = zmq.Context()
        socket = _context.socket(zmq.REP)
        socket.bind(_address)
        # -----------------------------------------
        stop = False
        ipc = _address.startswith('ipc://')
        while not(stop):  # pragma : no cover
            try:
                # Wait for next request from client
                raw = socket.recv()
                # if address is ipc, it means that the "server" process had been start from another process
                # -> the parent needs a way to stop imperatively the "server" process, thus the following line :
                if ipc:
                    if raw in ['stop', b'stop']:
                        socket.send_string('ok')
                        break
                # Handling request :
                data = fixbytes(msgpack.loads(raw, encoding='utf8'))
                obj = g
                for item in data:
                    method_name, args, kwargs = item
                    method = getattr(obj, method_name)
                    obj = method(*args, **kwargs)
                result = marshall(obj)
                socket.send(msgpack.dumps(result, encoding='utf8'))
            except KeyboardInterrupt:
                stop = True
            except:
                result = {
                    '__': 'x',
                    'd': traceback.format_exc()
                }
                try:
                    socket.send(msgpack.dumps(result, encoding='utf8'))
                except:
                    pass  # The server must not fail
                    # TODO : At least log an info
    finally:
        g.close()


def read_config_file(path):
    # defaults
    address = backend = script_import = None
    params = {}
    # parsing config file
    config = RawConfigParser()
    config.read(path)
    if config.has_section('server'):
        items = dict(config.items('server'))
        address = items.get('address', None)
        backend = items.get('backend', None)
    if config.has_section('scripts'):
        items = dict(config.items('scripts'))
        script_import = items.get('import', None)
    if config.has_section('backend'):
        params = dict(config.items('backend'))
    return address, backend, params, script_import


def parse_params(params):
    result = {}
    for param in params:
        key, value = param.split(':')
        result[key] = value
    return result


def main(cmds=None, verbose=True, context=None):

    # default config :
    default_address = "tcp://127.0.0.1:5555"
    default_backend = "grapheekdb.backends.data.localmem.LocalMemoryGraph"
    default_scripts = ''
    default_params  = {}

    # parsing command line :
    parser = argparse.ArgumentParser(description='Run GrapheekDB server.', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-a',
        '--address',
        dest='address',
        required=False,
        help="""The address string. This has the form 'protocol://interface:port',
        for example 'tcp://127.0.0.1:5555'. Protocols supported include
        tcp, inproc and ipc.
        """
    )
    parser.add_argument(
        "-b",
        "--backend",
        dest="backend",
        required=False,
        help='Data backend. must be of the form (for instance) : "grapheekdb.backends.data.kyotocab.KyotoCabinetGraph"',
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        required=False,
        help='Server configuration file',
    )
    parser.add_argument(
        "-s",
        "--scripts",
        dest="scripts",
        required=False,
        default=default_scripts,
        help='Server side scripts module path. Must be of the form : path.to.my_module1.scripts:path.to.my_module2.scripts:[.. additional module path ..]',
    )
    parser.add_argument(
        "params",
        nargs="*",
        default="",
        help='Data backend parameters. must be of the form : "key1:value1 key2:value ..."'
    )
    if cmds is not None:
        args = parser.parse_args(cmds)
    else:  # pragma : no cover
        args = parser.parse_args()
    # Building runserver arguments & keyword arguments
    # There's 3 steps to build them :
    # 1st step : use default values (done in first lines of main )
    # 2nd step : Read value from config file path passed as argument (if there's one). Override existing previous values
    # 3rd step :  Read value from command line options. Override existing previous values
    #
    # let's do the 2nd step : configuration file path
    config_address = config_backend = config_scripts = None
    config_params = {}
    line_config = args.config
    if line_config:  # pragma : no cover
        config_address, config_backend, config_params, config_scripts = read_config_file(line_config)
    # 3nd step : command line options :
    line_address = args.address
    line_backend = args.backend
    line_scripts = parse_params(args.scripts)
    line_params = parse_params(args.params)
    # Synthesis :
    address = line_address or config_address or default_address
    backend = line_backend or config_backend or default_backend
    scripts = line_scripts or config_scripts or default_scripts
    default_params.update(config_params)
    default_params.update(line_params)
    params = default_params
    config = {
        'version': __version__,
        'address': address,
        'backend': backend,
        'params': params,
        'scripts': scripts
    }
    label = "GrapheekDB Server version %(version)s" % config
    if verbose:  # pragma : no cover
        print(label)
        print('=' * len(label))
        print("Address              : %(address)s  <--  use the same address in client" % config)
        print("Backend              : %(backend)s" % config)
        print("Params               : %(params)s" % config)
        print("Server scripts paths : %(scripts)s" % config)
        print("Quit the server with CONTROL-C.")
    runserver(address, backend, scripts, context, **params)


if __name__ == '__main__':  # pragma : no cover
    main()



#!/usr/bin/env
# -*- coding: utf-8 -*-

from grapheekdb.lib.undef import UNDEFINED

from grapheekdb.backends.data.keys import METADATA_VERTEX_COUNTER, METADATA_EDGE_COUNTER
from grapheekdb.backends.data.keys import KIND_VERTEX
from grapheekdb.backends.data.keys import METADATA_EDGE_ID_LIST_PREFIX, METADATA_VERTEX_ID_LIST_PREFIX
from grapheekdb.backends.data.keys import CHUNK_SIZE
from grapheekdb.backends.data.keys import build_key


def choose_index_or_scan(seq_count, indexes, filters):
    """
    Helper function to choose between index use or sequential scan

    :param seq_count:
        an estimation of the sequential scan operation count (currently : number of entity_ids)
    :type seq_count:
        integer
    :param indexes:
        indexes that will challenge sequential scan. each index
        will be asked for an estimation of operation count
    :type indexes:
        BaseIndex child class instance

    :returns:
        the best index if one have been found with a probable best execution time than sequential scan
        OR None if no index was found
    """
    best_estimation = seq_count
    best_index = None
    for index in indexes:
        estimation = index.estimate(None, filters)
        if estimation < 0:  # normally : -1 (aka infinite)
            # special case : index is not competent to handle this filter
            continue
        if estimation < best_estimation:
            best_index = index
            best_estimation = estimation
    return best_index


class Optimizer(object):

    def __init__(self, graph):
        self._graph = graph

    def get_kind_ids(self, txn, kind):
        ENTITY_COUNTER = METADATA_VERTEX_COUNTER if kind == KIND_VERTEX else METADATA_EDGE_COUNTER
        METADATA_ID_LIST_PREFIX = METADATA_VERTEX_ID_LIST_PREFIX if kind == KIND_VERTEX else METADATA_EDGE_ID_LIST_PREFIX
        limit = int(self._graph._get(None, ENTITY_COUNTER)) // CHUNK_SIZE
        keys = [build_key(METADATA_ID_LIST_PREFIX, i) for i in range(0, limit + 1)]
        list_entity_ids = self._graph._bulk_get_lst(txn, keys)
        for entity_ids in list_entity_ids:
            if entity_ids != UNDEFINED:
                for entity_id in entity_ids:
                    yield entity_id

    def index_or_seq_scan_iterator(self, _kind, _seq_count, **filters):

        def indexed_entity_generator(entity_ids):
            for entity_id in entity_ids:
                yield entity_id

        iterator = None
        indexes = self._graph._node_indexes if _kind == KIND_VERTEX else self._graph._edge_indexes

        if filters and indexes:
            # There's some filters, so it *may* be useful to use filters
            # In order to know if it is useful, we will ask  every filter
            # to estimate the number of entity ids that they could return
            # we will, then, choose between the best index (given the filters)
            # and sequential iterator
            best_index = choose_index_or_scan(_seq_count, indexes, filters)
            if best_index:
                # Ok, an index will (hopefully :p) returns less ids than the seq scan, let's get ids :
                entity_ids = best_index.ids(None, filters)
                iterator = indexed_entity_generator(entity_ids)

        return iterator



# -*- coding:utf-8 -*-

import zmq
import msgpack

from grapheekdb.backends.data.keys import KIND_VERTEX, KIND_EDGE

from grapheekdb.lib.validations import check_valid_data

from grapheekdb.lib.readwrite import GraphReadWrite
from grapheekdb.lib.nx import GraphNx

from grapheekdb.lib.exceptions import GrapheekException
from grapheekdb.lib.exceptions import GrapheekDataException
from grapheekdb.lib.exceptions import GrapheekUnmarshallingException
from grapheekdb.lib.exceptions import GrapheekInvalidDataTypeException

try:  # pragma : no cover
    LONG = long
except NameError:  # pragma : no cover
    LONG = int
NONETYPE = type(None)
try:  # pragma : no cover
    UNICODE = unicode
except NameError:  # pragma : no cover
    UNICODE = str


def unmarshall(graph, item):
    if isinstance(item, (bool, float, int, str, LONG, NONETYPE, UNICODE)):
        return item
    if isinstance(item, bytes):  # pragma : no cover
        return str(item, encoding='utf8')
    if isinstance(item, (ProxyNode, ProxyEdge)):
        return item
    elif isinstance(item, dict):
        # Check if item is in fact a "special" dict (<-> handling an instance serialization)
        kind = item.get('__', None)
        if kind is not None:
            # If msg is an exception, raise it rapidly (do no count an caller to raise sth)
            # (But this doesn't forbid caller to catch exception, btw)
            if kind == 'x':
                raise GrapheekException(
                    '\n____________________________________\n*** SERVER *** exception traceback :\n%s'
                    % (item['d'],)
                )
            elif kind == 'n':
                return ProxyNode(graph, item['_i'], **unmarshall(graph, item['d']))
            elif kind == 'e':
                return ProxyEdge(graph, item['_i'], **unmarshall(graph, item['d']))
            elif kind == 'd':
                return dict((unmarshall(graph, key), unmarshall(graph, value)) for key, value in item['d'])
        else:
            items = [(unmarshall(graph, key), unmarshall(graph, value)) for key, value in list(item.items())]
            return dict(items)
    elif isinstance(item, (list, tuple)):
        return [unmarshall(graph, x) for x in item]
    raise GrapheekUnmarshallingException('Unknown type %s' % type(item))


def sanitize(item):
    if isinstance(item, (bool, float, int, str, LONG, NONETYPE, UNICODE)):
        return item
    if isinstance(item, bytes):  # pragma : no cover
        return str(item, encoding='utf8')
    if isinstance(item, dict):
        items = [(sanitize(key), sanitize(value)) for key, value in list(item.items())]
        return dict(items)
    elif isinstance(item, (list, tuple, set)):
        return [sanitize(x) for x in item]


def check_proxy_node(entity):
    if not(isinstance(entity, ProxyNode)):
        raise GrapheekInvalidDataTypeException('%s is not a node')


class ProxyDotExport(object):

    def __init__(self, dotstring):
        self._dotstring = dotstring

    def __repr__(self):
        return self._dotstring

    def to_dot(self):
        return self._dotstring


class ProxyIteratorCommon(object):

    def _request(self):
        return self._graph._request(self._commands)

    def __iter__(self):
        return self

    def __le__(self, it2):
        return self._graph.issubset(self, it2)

    def __ge__(self, it2):
        return self._graph.issuperset(self, it2)

    def __or__(self, it2):
        # | => union
        return self._graph.union(self, it2)

    def __and__(self, it2):
        # & => intersection
        return self._graph.intersection(self, it2)

    def __sub__(self, it2):
        # - => difference
        return self._graph.difference(self, it2)

    def __xor__(self, it2):
        # ^ => symmetric_difference
        return self._graph.symmetric_difference(self, it2)

    def next(self):
        # Python 2 compatibility
        return self.__next__()

    def __next__(self):
        if self._current_iteration is None:
            response = self._request()
            self._current_iteration = iter(response)
        try:
            return next(self._current_iteration)
        except StopIteration:
            self._current_iteration = None
        raise StopIteration

    def _do(self, _method, *args, **kwargs):
        command = [_method, args, kwargs]
        return self.__class__(self._graph, self._commands + [command])

    def _raw(self, _method, *args, **kwargs):
        command = [_method, args, kwargs]
        response = self._graph._request(self._commands + [command])
        return unmarshall(self._graph, response)


class ProxyEntityAggregate(ProxyIteratorCommon):

    def __init__(self, graph, commands):
        self._graph = graph
        self._commands = commands
        self._current_iteration = None

    def entities(self):
        command = ['entities', [], {}]
        return ProxyEntityIterator(self._graph, self._commands + [command])

    def limit(self, number):
        return self._do('limit', number)


class ProxyEntityIterator(ProxyIteratorCommon):

    def __init__(self, graph, commands):
        self._graph = graph
        self._commands = commands
        self._current_iteration = None

    def _jump(self, _traversal, *args, **kwargs):
        return self._do(_traversal, *args, **kwargs)

    def inV(self, *args, **kwargs):
        return self._jump('inV', *args, **kwargs)

    def outV(self, *args, **kwargs):
        return self._jump('outV', *args, **kwargs)

    def bothV(self, *args, **kwargs):
        return self._jump('bothV', *args, **kwargs)

    def in_(self, *args, **kwargs):
        return self._jump('in_', *args, **kwargs)

    def out_(self, *args, **kwargs):
        return self._jump('out_', *args, **kwargs)

    def both_(self, *args, **kwargs):
        return self._jump('both_', *args, **kwargs)

    def inE(self, **kwargs):
        # TODO : Check that the "ring" on which we are is composed of nodes (NOT edges)
        return self._jump('inE', **kwargs)

    def outE(self, **kwargs):
        # TODO : Check that the "ring" on which we are is composed of nodes (NOT edges)
        return self._jump('outE', **kwargs)

    def bothE(self, **kwargs):
        # TODO : Check that the "ring" on which we are is composed of nodes (NOT edges)
        return self._jump('bothE', **kwargs)

    def random(self, *args, **kwargs):
        return self._jump('random', *args, **kwargs)

    def dedup(self, *aliases):
        return self._do('dedup', *aliases)

    def without(self, *aliases):
        return self._do('without', *aliases)

    def data(self, *args, **kwargs):
        return self._raw('data', *args, **kwargs)

    def limit(self, count):
        return self._do('limit', count)

    def aka(self, alias):
        return self._do('aka', alias)

    def count(self):
        return int(self._raw('count'))

    def remove(self):
        self._raw('remove')

    def update(self, **updates):
        self._raw('update', **updates)

    def all(self):
        return self._raw('all')

    def ids(self):
        return self._raw('ids')

    def collect(self, *aliases):
        return self._raw('collect', *aliases)

    def sum(self, expr=None, asc=False):
        command = ('sum', [expr], dict(asc=asc))
        return ProxyEntityAggregate(self._graph, self._commands + [command])

    def percent(self, expr=None, asc=False):
        command = ('percent', [expr], dict(asc=asc))
        return ProxyEntityAggregate(self._graph, self._commands + [command])

    def call(self, _fname, *aliases, **params):
        return self._raw('call', _fname, *aliases, **params)

    def request(self, _fname, *aliases, **params):
        return self._raw('request', _fname, *aliases, **params)

    def dot(self, node_label=None, edge_label=None, limit=100):
        dotstring = self._raw('_dot_str', node_label, edge_label, limit)
        return ProxyDotExport(dotstring)


class ProxyEntity(object):

    def __init__(self, _graph, _entity_id, **data):
        check_valid_data(data)
        self._entity_id = _entity_id
        self._graph = _graph
        self._commands = [[self._kind, [_entity_id], {}]]
        self._data = data

    def __repr__(self):
        kind = 'node' if self._kind == KIND_VERTEX else 'edge'
        return '<%s id:%s data:%s>' % (kind, self._entity_id, repr(self._data))

    def __eq__(self, other):
        return other.__class__ == self.__class__ and other.get_id() == self.get_id() and other._graph == self._graph

    def __ne__(self, other):
        # Cf Python Doc : http://docs.python.org/2/reference/datamodel.html#object.__ne__
        # "When defining __eq__(), one should also define __ne__() so that the operators will behave as expected"
        return not self.__eq__(other)

    def __getattr__(self, attr):
        # Get value form local data
        if attr in self._data:
            return self._data[attr]
        # Try to update data :
        data = self._data = self.data()
        try:
            return data[attr]
        except KeyError:
            raise AttributeError("%r instance has no attribute %r" % (self.__class__, attr))

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            self.__dict__.update({attr: value})
        else:
            command = ['update_data', [self._kind, self._entity_id, attr, value], {}]
            self._graph._request([command])

    def _do(self, _method, *args, **kwargs):
        command = [_method, args, kwargs]
        return ProxyEntityIterator(self._graph, self._commands + [command])

    def _raw(self, _method, *args, **kwargs):
        command = [_method, args, kwargs]
        response = self._graph._request(self._commands + [command])
        return unmarshall(self._graph, response)

    def get(self, field, default=None):
        try:
            return self.__getattr__(field)
        except AttributeError:
            return default

    def inV(self, *args, **kwargs):
        return self._do('inV', *args, **kwargs)

    def outV(self, *args, **kwargs):
        return self._do('outV', *args, **kwargs)

    def bothV(self, *args, **kwargs):
        return self._do('bothV', *args, **kwargs)

    def in_(self, *args, **kwargs):
        return self._do('in_', *args, **kwargs)

    def out_(self, *args, **kwargs):
        return self._do('out_', *args, **kwargs)

    def both_(self, *args, **kwargs):
        return self._do('both_', *args, **kwargs)

    def random(self, *args, **kwargs):
        return self._do('random', *args, **kwargs)

    def data(self, *args, **kwargs):
        return self._raw('data', *args, **kwargs)

    def get_id(self):
        return int(self._entity_id)

    def remove(self):
        return self._raw('remove')

    def update(self, **updates):
        self._raw('update', **updates)
        # No exception raised -> Apply updates on self.__dict__
        self.__dict__.update(updates)

    def aka(self, alias):
        return self._do('aka', alias)


class ProxyNode(ProxyEntity):

    def __init__(self, _graph, _node_id, **data):
        self._kind = KIND_VERTEX
        super(ProxyNode, self).__init__(_graph, _node_id, **data)

    def inE(self, **kwargs):
        return self._do('inE', **kwargs)

    def outE(self, **kwargs):
        return self._do('outE', **kwargs)

    def bothE(self, **kwargs):
        return self._do('bothE', **kwargs)


class ProxyEdge(ProxyEntity):

    def __init__(self, _graph, _edge_id, **data):
        self._kind = KIND_EDGE
        super(ProxyEdge, self).__init__(_graph, _edge_id, **data)


class ProxyGraph(GraphReadWrite, GraphNx):

    def __init__(self, address):  # pragma : no cover
        self._address = address
        #self._zmq_context = zmq.Context()
        #self._zmq_socket = self._zmq_context.socket(zmq.REQ)
        #self._zmq_socket.connect(address)

    def _request(self, commands):
        data = msgpack.dumps(sanitize(commands), encoding='utf8')
        zmq_context = zmq.Context()
        zmq_socket = zmq_context.socket(zmq.REQ)
        zmq_socket.connect(self._address)
        zmq_socket.send(data)
        raw = zmq_socket.recv()
        msg = msgpack.loads(raw, encoding='utf8')
        return unmarshall(self, msg)

    def v(self, entity_id):
        return ProxyNode(self, entity_id)

    def V(self, **kwargs):
        command = [KIND_VERTEX.upper(), [], kwargs]
        return ProxyEntityIterator(self, [command])

    def e(self, entity_id):
        return ProxyEdge(self, entity_id)

    def E(self, **kwargs):
        command = [KIND_EDGE.upper(), [], kwargs]
        return ProxyEntityIterator(self, [command])

    def add_node(self, **kwargs):
        check_valid_data(kwargs)
        command = ['add_node', [], kwargs]
        res = self._request([command])
        return res

    def bulk_add_node(self, node_defns, silent=False):
        for data in node_defns:
            check_valid_data(data)
        command = ['bulk_add_node', [node_defns], {'silent': silent}]
        return self._request([command])

    def add_edge(self, _source, _target, **kwargs):
        check_valid_data(kwargs)
        check_proxy_node(_source)
        check_proxy_node(_target)
        source_id = _source.get_id()
        target_id = _target.get_id()
        command = ['add_edge_by_ids', [source_id, target_id], kwargs]
        return self._request([command])

    def bulk_add_edge(self, edge_defns, silent=False):
        try:
            for source, target, data in edge_defns:
                check_proxy_node(source)
                check_proxy_node(target)
                check_valid_data(data)
        except (TypeError, ValueError) as e:
            raise GrapheekDataException(repr(e))
        # Don't be suprised by [[ ... ]] instead of [ ... ]
        # Server function is expecting only one argument :
        args = [[(source.get_id(), target.get_id(), data) for source, target, data in edge_defns]]
        command = ['bulk_add_edge_by_ids', args, {'silent': silent}]
        return self._request([command])

    def _add_entity_index(self, _kind, *fields, **filters):
        assert(_kind in (KIND_EDGE, KIND_VERTEX))
        command = ['add_node_index' if _kind == KIND_VERTEX else 'add_edge_index', fields, filters]
        return self._request([command])

    def add_node_index(self, *fields, **filters):
        return self._add_entity_index(KIND_VERTEX, *fields, **filters)

    def add_edge_index(self, *fields, **filters):
        return self._add_entity_index(KIND_EDGE, *fields, **filters)

    def get_node_indexes(self):
        command = ['get_node_indexes', [], {}]
        return self._request([command])

    def get_edge_indexes(self):
        command = ['get_edge_indexes', [], {}]
        return self._request([command])

    def _remove_entity_index(self, _kind, *fields, **filters):
        assert(_kind in (KIND_EDGE, KIND_VERTEX))
        command = ['remove_node_index' if _kind == KIND_VERTEX else 'remove_edge_index', fields, filters]
        return self._request([command])

    def remove_node_index(self, *fields, **filters):
        return self._remove_entity_index(KIND_VERTEX, *fields, **filters)

    def remove_edge_index(self, *fields, **filters):
        return self._remove_entity_index(KIND_EDGE, *fields, **filters)

    def _operation_helper(self, operation, *entity_iterators):
        command = [operation, [entity_iterator._commands for entity_iterator in entity_iterators], {}]
        return ProxyEntityIterator(self, [command])

    def issubset(self, *entity_iterators):
        command = ['_marshalled_issubset', [entity_iterator._commands for entity_iterator in entity_iterators], {}]
        return self._request([command])

    def issuperset(self, *entity_iterators):
        command = ['_marshalled_issuperset', [entity_iterator._commands for entity_iterator in entity_iterators], {}]
        return self._request([command])

    def union(self, *entity_iterators):
        return self._operation_helper('_marshalled_union', *entity_iterators)

    def intersection(self, *entity_iterators):
        return self._operation_helper('_marshalled_intersection', *entity_iterators)

    def difference(self, *entity_iterators):
        return self._operation_helper('_marshalled_difference', *entity_iterators)

    def symmetric_difference(self, *entity_iterators):
        return self._operation_helper('_marshalled_symmetric_difference', *entity_iterators)



# -*- coding:utf-8 -*-
"""
This module is mainly used for tests
It exposes a single function that echoes arguments and keyword arguments
"""


def echo(g, _ctxs, *args, **kwargs):
    return [dict(args=args, kwargs=kwargs) for ctx in _ctxs]



# -*- coding:utf-8 -*-


def lookup_exact(reference_value, field_value):
    return field_value == reference_value


def lookup_iexact(reference_value, field_value):
    return lookup_exact(field_value.lower(), reference_value.lower())


def lookup_contains(reference_value, field_value):
    return reference_value in field_value


def lookup_icontains(reference_value, field_value):
    return reference_value.lower() in field_value.lower()


def lookup_in(reference_value, field_value):
    return field_value in reference_value


def lookup_gt(reference_value, field_value):
    return field_value > reference_value


def lookup_gte(reference_value, field_value):
    return field_value >= reference_value


def lookup_lt(reference_value, field_value):
    return field_value < reference_value


def lookup_lte(reference_value, field_value):
    return field_value <= reference_value


def lookup_startswith(reference_value, field_value):
    return field_value.startswith(reference_value)


def lookup_istartswith(reference_value, field_value):
    return lookup_startswith(reference_value.lower(), field_value.lower())


def lookup_endswith(reference_value, field_value):
    return field_value.endswith(reference_value)


def lookup_iendswith(reference_value, field_value):
    return lookup_endswith(reference_value.lower(), field_value.lower())


def lookup_isnull(reference_value, field_value):
    return (field_value is None) == reference_value



# -*- coding:utf-8 -*-

from grapheekdb.backends.data.base import BaseGraph

from grapheekdb.lib.undef import UNDEFINED


class LocalMemoryGraph(BaseGraph):

    def __init__(self):
        self._dic = {}
        super(LocalMemoryGraph, self).__init__()
        self._ensure_prepared()
        self._closed = False

    # Start method overriding :

    def _db_close(self):  # pragma : no cover
        pass

    def _transaction_begin(self):
        return True

    def _transaction_commit(self, txn):
        pass

    def _transaction_rollback(self, txn):
        pass

    def _has_key(self, key):
        return key in self._dic

    def _get(self, txn, key):
        return self._dic.get(key, UNDEFINED)  # Not returning None, as None is a valid value

    def _bulk_get(self, txn, keys):
        dic = {}
        for key in keys:
            value = self._get(txn, key)
            if value != UNDEFINED:
                dic[key] = value
        return dic

    def _set(self, txn, key, value):
        self._dic[key] = value

    def _bulk_set(self, txn, updates):
        self._dic.update(updates)

    def _remove(self, txn, key):
        try:
            del self._dic[key]
        except KeyError:
            pass  # I want _remove method to be idempotent

    def _bulk_remove(self, txn, keys):
        for key in keys:
            try:
                del self._dic[key]
            except KeyError:  # pragma : no cover
                pass  # I want _remove method to be idempotent

    def _remove_prefix(self, txn, prefix):
        remove_keys = [key for key in self._dic if key.startswith(prefix)]
        for key in remove_keys:
            self._remove(txn, key)

    def _append_to_lst(self, txn, key, value):
        lst = self._get(txn, key)
        if lst == UNDEFINED:
            lst = []
            lst.append(value)
            self._set(txn, key, lst)
        else:
            lst.append(value)

    def _bulk_append_to_lst(self, txn, key, value):
        lst = self._get(txn, key)
        if lst == UNDEFINED:
            lst = []
            lst.extend(value)
            self._set(txn, key, lst)
        else:
            lst.extend(value)




#!/usr/bin/env
# -*- coding: utf-8 -*-

try:  # pragma : no cover
    from itertools import izip_longest as zip_longest
except ImportError:  # pragma : no cover
    from itertools import zip_longest

from itertools import product
from collections import defaultdict

import json

from grapheekdb.lib.undef import UNDEFINED

from grapheekdb.backends.data.keys import FORBIDDEN_KEY
from grapheekdb.backends.data.keys import CHUNK_SIZE
from grapheekdb.backends.data.keys import COUNT_SUFFIX
from grapheekdb.backends.data.keys import DATA_SUFFIX
from grapheekdb.backends.data.keys import build_key

from grapheekdb.backends.data.filtertools import get_exact_filters, build_filter_funcs, entity_match

from grapheekdb.lib.exceptions import GrapheekIncompetentIndexException


def normalize_value(obj):
    return json.dumps(obj)


class BaseIndex(object):

    def __init__(self, _graph, _kind, _prefix):
        self._graph = _graph
        self._kind = _kind
        self._prefix = _prefix

    def bulk_add(self, txn, id_iterator):
        raise NotImplementedError

    def add(self, txn, entity_id, data):
        raise NotImplementedError

    def remove(self, txn, entity, data):
        raise NotImplementedError

    def estimate(self, txn, filters):
        raise NotImplementedError

    def ids(self, txn, filters):
        raise NotImplementedError


class ExactIndex(BaseIndex):

    def __init__(self, _graph, _kind, _prefix, *fields, **filters):
        super(ExactIndex, self).__init__(_graph, _kind, _prefix)
        self._fields = list(fields)
        self._set_fields = set(fields)
        self._filters = filters
        self._filter_items = set(filters.items())
        self._set_filter_keys = set(filters.keys())
        self._filter_length = len(filters)
        self._filter_funcs = build_filter_funcs(**filters)
        self._fields.sort()

    def delete(self, txn):
        """
        delete the index
        """
        self._graph._remove_prefix(txn, self._prefix)

    def bulk_add(self, txn, id_iterator):

        def save_lst(txn, chunk_idx, value, lst):
            lst_key = build_key(self._prefix, chunk_idx, value)
            self._graph._set_lst(txn, lst_key, lst)

        kind = self._kind
        values_to_id = defaultdict(list)
        # Getting data for all entity_id in chunk of <CHUNK_SIZE>
        filter_length = self._filter_length
        filter_funcs = self._filter_funcs
        for entity_ids in zip_longest(*([id_iterator] * CHUNK_SIZE), fillvalue=FORBIDDEN_KEY):
            entity_datas = self._graph._bulk_get(txn, [build_key(kind, entity_id, DATA_SUFFIX) for entity_id in entity_ids if entity_id != FORBIDDEN_KEY])
            for entity_key, data in list(entity_datas.items()):
                if filter_length and not entity_match(filter_funcs, data):
                    continue
                entity_id = int(entity_key.replace(kind + '/', '').replace('/' + DATA_SUFFIX, ''))  # not very clean :(
                lst = [data.get(field, None) for field in self._fields]
                # Completing values_to_id (it's here that the reversal value -> entity_id become possible)
                values_to_id[normalize_value(lst)].append(entity_id)
        # This dic will contain the mapping entity id index key -> value chunk id
        entity_dict = {}
        for value, entity_ids in values_to_id.items():
            lst = []
            value_count_key = build_key(self._prefix, COUNT_SUFFIX, value)
            value_count = self._graph._get(txn, value_count_key)
            if value_count == UNDEFINED:
                value_count = 0
            chunk_idx = value_count // CHUNK_SIZE
            for entity_id in entity_ids:
                value_count += 1
                entity_dict[build_key(self._prefix, entity_id)] = build_key(chunk_idx, value)  # Need to keep an info of relation between entity and the chunk id that contains it
                lst.append(entity_id)
                if value_count % CHUNK_SIZE == 0:
                    save_lst(txn, chunk_idx, value, lst)
                    lst = []
                    chunk_idx += 1
            if lst:
                save_lst(txn, chunk_idx, value, lst)
            self._graph._set(txn, value_count_key, value_count)
        self._graph._bulk_set(txn, entity_dict)

    def add(self, txn, entity_id, data):
        if self._filter_length and not entity_match(self._filter_funcs, data):
            return
        value = normalize_value([data.get(field, None) for field in self._fields])
        # Get current chunk idx for this value (NOTE : value is a string representing a list)
        value_count_key = build_key(self._prefix, COUNT_SUFFIX, value)
        value_count = 0
        value_count_current = self._graph._get(txn, value_count_key)
        if value_count_current != UNDEFINED:
            value_count = int(value_count_current)
        # Incrementing entity count for value
        self._graph._set(txn, value_count_key, value_count + 1)
        #  1st adding entity_id to good chunk :
        chunk_idx = value_count // CHUNK_SIZE
        chunk_key = build_key(self._prefix, chunk_idx, value)
        self._graph._append_to_lst(txn, chunk_key, entity_id)
        # Say that entity_id is in a chunk
        # (we need to keep chunk idx, so that entity id can be removed fastly
        # from chunk when entity will be removed from DB)
        self._graph._set(txn, build_key(self._prefix, entity_id), build_key(chunk_idx, value))

    def remove(self, txn, entity_id):
        entity_chunk_sub_key = self._graph._get(txn, build_key(self._prefix, entity_id))
        if entity_chunk_sub_key != UNDEFINED:
            # remove entity_id from chunk
            chunk_key = build_key(self._prefix, entity_chunk_sub_key)
            self._graph._remove_from_lst(txn, chunk_key, entity_id)
            # Finally remove entity_chunk_sub_key
            self._graph._remove(txn, entity_chunk_sub_key)

    def _compatible_filters(self, filters):
        # FIXME : This code works but is hard to understand (for me)
        # And it expects get_exact_filters to build a special value (a list) for __in lookups
        # This method should be as short as :
        # index.supersedes(filters)
        # 1st -a- check that filters is compatible with (partial) index filters
        exact_filters = get_exact_filters(**filters)
        set_exact_filter_keys = set(exact_filters.keys())
        if not(self._set_fields.issubset(set_exact_filter_keys)):
            return False
        if not(self._set_filter_keys.issubset(set(exact_filters.keys()))):
            return False
        common_filters = {}
        if self._filter_length:
            for key, value in list(self._filters.items()):
                out_value = exact_filters[key]
                if value == out_value or (isinstance(out_value, list) and value in out_value):
                    common_filters[key] = value
                else:
                    return False
        exact_filters.update(common_filters)
        # 2nd, check on exact filters
        resulting_filters = []
        for field in self._fields:
            v = exact_filters[field]
            resulting_filters.append(v if isinstance(v, list) else [v])
        return resulting_filters

    def estimate(self, txn, filters):
        resulting_filters = self._compatible_filters(filters)
        if resulting_filters is False:
            return -1  # -1 says that this index shouldn't be used
        total = 0
        for values in product(*resulting_filters):
            value = normalize_value(values)
            value_count_key = build_key(self._prefix, COUNT_SUFFIX, value)
            value_count_current = self._graph._get(txn, value_count_key)
            if value_count_current != UNDEFINED:
                total += int(value_count_current)
        return total  # note : this is an over estimation

    def ids(self, txn, filters):
        resulting_filters = self._compatible_filters(filters)
        if resulting_filters is False:
            raise GrapheekIncompetentIndexException("This index shouldn't have been used")
        for values in product(*resulting_filters):
            value = normalize_value(values)
            value_count_key = build_key(self._prefix, COUNT_SUFFIX, value)
            value_count_current = self._graph._get(txn, value_count_key)
            if value_count_current != UNDEFINED:
                for chunk_idx in range(0, 1 + int(value_count_current) // CHUNK_SIZE):
                    entity_ids_key = build_key(self._prefix, chunk_idx, value)
                    entity_ids = self._graph._get_lst(txn, entity_ids_key)
                    if entity_ids != UNDEFINED:
                        for entity_id in entity_ids:
                            yield entity_id



# -*- coding: utf-8 -*-
"""
Adapted from NetworkX algorithms to grapheekdb methods
"""

import heapq

from grapheekdb.backends.data.keys import KIND_VERTEX
from grapheekdb.lib.exceptions import GrapheekUnreachableNode


def single_source_dijkstra(g, src, tgt=None, cutoff=None, weight=None):
    """
    Compute shortest paths and lengths in a weighted graph g.
    Uses Dijkstra's algorithm for shortest paths.

    :param g:
        GrapheekDB graph

    :param src:
        starting node for path

    :type src:
        node

    :param tgt:
        ending node for path (optionnal)

    :type tgt:
        node

    :param cutoff:
        Depth to stop the search. Only paths of length <= cutoff are returned.

    :type cutoff:
        integer or float, optional

    :param weight:
        edge attribute that contains the weight

    :type weight:
        string


    :returns
        distance,path : dictionaries
        Returns a tuple of two dictionaries keyed by node id.
        The first dictionary stores distance from the source.
        The second stores the path from the source to that node.

    Notes
    ---------
    Edge weight attributes must be numerical.
    Distances are calculated as sums of weighted edges traversed.

    Based on the Python cookbook recipe (119466) at
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/119466

    This algorithm is not guaranteed to work if edge weights
    are negative or are floating point numbers
    (overflows and roundoff errors can cause problems).
    """
    source_id = src.get_id()
    target_id = tgt.get_id() if tgt is not None else None
    if src == tgt:
        return ({source_id: 0}, {source_id: [source_id]})
    dists = {}  # dictionary of final distances
    paths = {source_id: [source_id]}  # dictionary of paths
    seen = {source_id: 0}
    fringe = []  # use heapq with (distance,node_id) tuples
    heapq.heappush(fringe, (0, source_id))
    while fringe:
        (d, node_id) = heapq.heappop(fringe)
        if node_id in dists:
            continue  # already searched this node.
        dists[node_id] = d
        if node_id == target_id:
            break
        node = g._item_from_id(KIND_VERTEX, node_id)
        out_edges = node.outE()
        out_node_ids = node.outV()._iterate()

        for out_edge, out_node_id in zip(out_edges, out_node_ids):
            if weight is not None:
                edgedata = out_edge.data()
                vw_dist = d + edgedata[weight]
            else:
                vw_dist = d + 1
            if cutoff is not None:
                if vw_dist > cutoff:
                    continue
            if out_node_id in dists:
                if vw_dist < dists[out_node_id]:
                    raise ValueError('Contradictory paths found:',
                                     'negative weights?')
            elif out_node_id not in seen or vw_dist < seen[out_node_id]:
                seen[out_node_id] = vw_dist
                heapq.heappush(fringe, (vw_dist, out_node_id))
                paths[out_node_id] = paths[node_id] + [out_node_id]
    return (dists, paths)


def dijkstra_path(g, src, tgt, cutoff=None, weight=None):
    dists, paths = single_source_dijkstra(g, src, tgt, cutoff=cutoff, weight=weight)
    try:
        dist = dists[tgt.get_id()]
    except KeyError:
        raise GrapheekUnreachableNode("Unreachable node")
    path = []
    for node_id in paths[tgt.get_id()]:
        path.append(g._item_from_id(KIND_VERTEX, node_id))
    return (dist, path)





# -*- coding:utf-8 -*-

import sys

import lmdb
import msgpack

from grapheekdb.backends.data.base import BaseGraph

from grapheekdb.lib.undef import UNDEFINED

PYTHON3 = sys.version_info.major == 3

class LmdbGraph(BaseGraph):

    def __init__(self, path, map_size=1024 * 1024):
        # create the database object
        self._path = path
        # open the database
        self._env = lmdb.open(path, map_size=map_size)
        super(LmdbGraph, self).__init__()
        self._ensure_prepared()
        self._closed = False

    # Start method overriding :

    def _db_close(self):
        if not self._closed:
            self._env.close()

    def _transaction_begin(self):
        return self._env.begin(write=True)

    def _transaction_commit(self, txn):
        txn.commit()

    def _transaction_rollback(self, txn):
        txn.abort()

    def _has_key(self, key):
        k = key
        if PYTHON3:  # pragma : no cover
            k = bytes(k, encoding='utf8')
        with self._env.begin() as txn:
            return txn.get(k, UNDEFINED) != UNDEFINED

    def _get(self, txn, key):
        k = key
        if PYTHON3:  # pragma : no cover
            k = bytes(k, encoding='utf8')
        if txn is not None:
            raw_data = txn.get(k, UNDEFINED)
            if raw_data == UNDEFINED:
                return UNDEFINED
            res = msgpack.loads(raw_data, encoding='utf8')
            return res
        else:
            with self._env.begin() as txn:
                raw_data = txn.get(k, UNDEFINED)
                if raw_data == UNDEFINED:
                    return UNDEFINED
                res = msgpack.loads(raw_data, encoding='utf8')
                return res

    def _bulk_get(self, txn, keys):
        dic = {}
        for key in keys:
            value = self._get(txn, key)
            if value != UNDEFINED:
                dic[key] = value
        return dic

    def _set(self, txn, key, value):
        k = key
        if PYTHON3:  # pragma : no cover
            k = bytes(k, encoding='utf8')
        txn.put(k, msgpack.dumps(value, encoding='utf8'))

    def _bulk_set(self, txn, updates):
        for key, value in list(updates.items()):
            self._set(txn, key, value)

    def _remove(self, txn, key):
        k = key
        if PYTHON3:  # pragma : no cover
            k = bytes(k, encoding='utf8')
        txn.delete(k)

    def _bulk_remove(self, txn, keys):
        for key in keys:
            self._remove(txn, key)

    def _remove_prefix(self, txn, prefix):
        cursor = txn.cursor()
        for key, _ in cursor:
            if PYTHON3:  # pragma : no cover
                key = str(key, encoding='utf8')
            if key.startswith(prefix):
                cursor.delete()



# -*- coding: utf-8 -*-

from heapq import heappop, heappush

from grapheekdb.backends.data.keys import KIND_VERTEX
from grapheekdb.lib.exceptions import GrapheekUnreachableNode


def astar_path(g, src, tgt, heuristic=None, weight=None):
    """Return a list of nodes in a shortest path between source and target
    using the A* ("A-star") algorithm.

    There may be more than one shortest path.  This returns only one.

    :param g:
        GrapheekDB graph

    :param src:
        starting node for path

    :type src:
        node

    :param tgt:
        ending node for path (optionnal)

    :type tgt:
        node

    :param heuristic:
       A function to evaluate the estimate of the distance
       from the a node to the target.  The function takes
       two nodes arguments and must return a number.

    :type heuristic:
        function

    :param weight:
        edge attribute that contains the weight

    :type weight:
        string

    :returns
        distance,path : dictionaries
        Returns a tuple of two dictionaries keyed by node id.
        The first dictionary stores distance from the source.
        The second stores the path from the source to that node.

    """
    if heuristic is None:
        # The default heuristic is h=0 - same as Dijkstra's algorithm
        def heuristic(u, v):
            return 0
    # The queue stores priority, node, cost to reach, and parent.
    # Uses Python heapq to keep in priority order.
    # Add each node's hash to the queue to prevent the underlying heap from
    # attempting to compare the nodes themselves. The hash breaks ties in the
    # priority and is guarenteed unique for all nodes in the graph.
    source_id = src.get_id()
    target_id = tgt.get_id()
    target_data = tgt.data()
    queue = [(0, hash(source_id), source_id, 0, None)]

    # Maps enqueued nodes to distance of discovered paths and the
    # computed heuristics to target. We avoid computing the heuristics
    # more than once and inserting the node into the queue too many times.
    enqueued = {}
    # Maps explored nodes to parent closest to the source.
    explored = {}

    while queue:
        # Pop the smallest item from queue.
        _, __, curnode_id, dist, parent_id = heappop(queue)
        if curnode_id == target_id:
            path = [curnode_id]
            node_id = parent_id
            while node_id is not None:
                path.append(node_id)
                node_id = explored[node_id]
            path.reverse()
            return [g._item_from_id(KIND_VERTEX, idx) for idx in path]

        if curnode_id in explored:
            continue

        explored[curnode_id] = parent_id

        curnode = g._item_from_id(KIND_VERTEX, curnode_id)
        out_edges = curnode.outE()
        out_nodes = curnode.outV()

        for out_edge, out_node in zip(out_edges, out_nodes):
            out_node_id = out_node.get_id()
            if out_node_id in explored:
                continue
            if weight is not None:
                edgedata = out_edge.data()
                ncost = dist + edgedata[weight]
            else:
                ncost = dist + 1
            if out_node_id in enqueued:
                qcost, h = enqueued[out_node_id]
                # if qcost < ncost, a longer path to neighbor remains
                # enqueued. Removing it would need to filter the whole
                # queue, it's better just to leave it there and ignore
                # it when we visit the node a second time.
                if qcost <= ncost:
                    continue
            else:
                h = heuristic(out_node.data(), target_data)
            enqueued[out_node_id] = ncost, h
            heappush(queue, (ncost + h, hash(out_node_id), out_node_id,
                             ncost, curnode_id))

    raise GrapheekUnreachableNode("Node %s not reachable from %s" % (src, tgt))


def astar_path_length(g, src, tgt, heuristic=None, weight=None):
    """
    :param g:
        GrapheekDB graph

    :param src:
        starting node for path

    :type src:
        node

    :param tgt:
        ending node for path (optionnal)

    :type tgt:
        node

    :param heuristic:
       A function to evaluate the estimate of the distance
       from the a node to the target.  The function takes
       two nodes arguments and must return a number.

    :type heuristic:
        function

    :param weight:
        edge attribute that contains the weight

    :type weight:
        string

    :returns
        distance,path : dictionaries
        Return the length of the shortest path between source and target using
        the A* ("A-star") algorithm.
    """
    path = astar_path(g, src, tgt, heuristic, weight)
    result = 0
    # FIXME : Following implementation is quite poor from a performance point of view
    for source, target in zip(path[:-1], path[1:]):
        for out_edge in source.outE():
            if target in list(out_edge.outV()):
                result += out_edge.data().get(weight, 1)
                break
    return result



import ast
import operator as op
import re

from grapheekdb.lib.exceptions import GrapheekInvalidExpression

# supported operators
operators = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
             ast.Div: op.truediv, ast.FloorDiv: op.floordiv, ast.Pow: op.pow}
cmps = {ast.Lt: op.lt, ast.LtE: op.le, ast.Gt: op.gt, ast.GtE: op.ge, ast.Eq: op.eq}

# valid attribute name regex :
VALID_ATTRIBUTE = r'''^[a-zA-Z][a-zA-Z0-9_]*$'''
VALID_ATTRIBUTE_REGEX = re.compile(VALID_ATTRIBUTE)


def eval_expr(context, expr):
    try:
        return eval_(context, ast.parse(expr).body[0].value)
    except GrapheekInvalidExpression:
        raise
    except:
        raise GrapheekInvalidExpression


def eval_(context, node):
    if isinstance(node, ast.Num):
        return node.n
    elif isinstance(node, ast.operator):
        return operators[type(node)]
    elif isinstance(node, ast.cmpop):
        return cmps[type(node)]
    elif isinstance(node, ast.IfExp):
        return eval_(context, node.body) if eval_(context, node.test) else eval_(context, node.orelse)
    elif isinstance(node, ast.BinOp):
        return eval_(context, node.op)(eval_(context, node.left), eval_(context, node.right))
    elif isinstance(node, ast.Name):
        return context[node.id]
    elif isinstance(node, ast.Attribute):
        attr = node.attr
        if VALID_ATTRIBUTE_REGEX.match(attr):
            return eval_(context, node.value).get(node.attr, 0)
        raise GrapheekInvalidExpression("attributes must be of the form %s" % (VALID_ATTRIBUTE,))
    elif isinstance(node, ast.Compare):
        result = eval_(context, node.left)
        for op, comp in zip(node.ops, node.comparators):
            result = eval_(context, op)(result, eval_(context, comp))
        return result
    else:
        raise GrapheekInvalidExpression(repr(node))



# -*- coding:utf-8 -*-


class GraphNx(object):

    def nx_graph(self):
        # This may raise an exception
        # let's it raise to warn user that
        # networkx should be installed for this method to be used
        import networkx as nx
        G = nx.DiGraph()
        node_ids = set()
        edge_ids = set()
        for node in self.V():
            node_id = node.get_id()
            if not node_id in node_ids:
                G.add_node(node_id, **node.data())
                node_ids.add(node_id)
            for out_edge, out_node in zip(node.outE(), node.outV()):
                out_edge_id = out_edge.get_id()
                out_node_id = out_node.get_id()
                if not out_node_id in node_ids:
                    G.add_node(out_node_id, **out_node.data())
                    node_ids.add(out_node_id)
                if not out_edge_id in edge_ids:
                    G.add_edge(node_id, out_node_id, **out_edge.data())
                    edge_ids.add(out_edge_id)
        return G



# -*- coding:utf-8 -*-

from functools import partial
from itertools import tee

from . import lookups

from grapheekdb.lib.undef import UNDEFINED

from grapheekdb.backends.data.keys import DATA_SUFFIX

from grapheekdb.lib.exceptions import GrapheekSubLookupNotImplementedException
from grapheekdb.lib.exceptions import GrapheekInvalidLookupException


def get_exact_filters(**filters):
    exact_filters = {}
    for key, value in list(filters.items()):
        # Note : not only looking for exact lookups
        lst = key.split('__')
        len_lst = len(lst)
        if len_lst == 1:
            field = lst[0]
            clause = 'exact'
            exact_filters[field] = value
        elif len_lst == 2:
            field, clause = lst
            if clause in ['exact', 'in']:
                exact_filters[field] = value
        else:
            raise GrapheekSubLookupNotImplementedException
    return exact_filters


def build_filter_funcs(**filters):
    field = clause = None
    filter_funcs = []
    for key, value in list(filters.items()):
        lst = key.split('__')
        len_lst = len(lst)
        if len_lst == 1:
            field = lst[0]
            clause = 'exact'
            lookup_func = lookups.lookup_exact
        elif len_lst == 2:
            field, clause = lst
            try:
                lookup_func = getattr(lookups, 'lookup_' + lst[1])
            except AttributeError:
                raise GrapheekInvalidLookupException
        else:
            raise GrapheekSubLookupNotImplementedException
        filter_funcs.append((field, clause, partial(lookup_func, value)))
    return filter_funcs


def entity_match(filter_funcs, data, internals=None):
    if data == UNDEFINED:  # pragma : no cover
        # This could happen because of concurrent modifications (my different process or different threads)
        # This won't normally occur when load testing with greenlets -> thus no cover
        # (or worst : direct access to backend by another "process")
        return False
    for field, clause, filter_func in filter_funcs:
        try:
            field_value = data[field]
        except KeyError:
            try:
                field_value = internals[field]
            except (KeyError, TypeError):
                field_value = None
        if field_value is None and clause != 'isnull':
            return False
        try:
            match = filter_func(field_value)
        except:
            # Whatever happens, it shouldn't
            # So, let's consider that data wasn't ok for the filter
            return False
        if not(match):
            return False
    return True


def filter_entities(graph, kind, iterator, filter_funcs, on_item=False):
    _, cloned_iterator = tee(iterator)
    if filter_funcs:
        if on_item:
            for entity_id in cloned_iterator:
                data = graph._get(None, kind + '/' + str(entity_id) + '/' + DATA_SUFFIX)
                match = entity_match(filter_funcs, data, {'_id': entity_id})
                if match:
                    yield graph._item_from_id(kind, entity_id)
        else:
            for entity_id in cloned_iterator:
                data = graph._get(None, kind + '/' + str(entity_id) + '/' + DATA_SUFFIX)
                match = entity_match(filter_funcs, data, {'_id': entity_id})
                if match:
                    yield entity_id
    else:
        # I don't do : yield item if on_item else entity_id
        # Because item is not computed (when on_item is False, this is not useful to instantiate an Entity)
        if on_item:
            for entity_id in cloned_iterator:
                yield graph._item_from_id(kind, entity_id)
        else:
            for entity_id in cloned_iterator:
                yield entity_id



# -*- coding:utf-8 -*-

"""
This module contains keys pattern that are used to turn a KVS (Key Value Store) into a Graph Database
And various utilities related to keys

Here is the keys tree :
-----------------------

.
├── init                      INITIALIZED ?
├── xxxx                      Forbidden key
├── m                         METADATA
│   ├── e                     edge
│   │   ├── c                 (m/e/c)  edges counter (it will never decrease as it will serve to get new id)
│   │   ├── r                 (m/e/r)  removed edges counter
│   │   ├── id                edge ids list
│   │   │   └── <id % 1000>   (m/e/id/<id % 1000>) a list of max 1000 edge ids (helps to speed id iteration)
│   │   ├── ci                (m/e/ci) edges counter for index
│   │   ├── il                (m/e/il) edge indexes list
│   │   ├── if                (m/e/if) edge indexes fields
│   │   │   └── <id>          (m/e/i/<id>) edge index id fields
│   │   └── i                 index
│   │       └── *key sorted   (m/e/i/*sorted_key)  indexes
│   └── v                     vertex (aka node)
│       ├── c                 (m/v/c)  vertices counter (it will never decrease as it will serve to get new id)
│       ├── r                 (m/v/r)  removed vertices counter
│       ├── id                vertex ids list
│       │   └── <id % 1000>   (m/e/id/<id % 1000>) a list of max 1000 vertex ids (helps to speed id iteration)
│       ├── ci                (m/v/ci) vertices counter for index
│       ├── il                (m/v/il) vertices indexes list
│       ├── if                (m/e/if) vertices indexes fields
│       │   └── <id>          (m/e/i/<id>) vertex index id fields
│       └── i                 index
│           └── *key sorted   (m/v/i/*sorted_key)  indexes
├── i                         INDEXES
│   ├── e                     edge
│   │   └── <id>              edge index id
│   │       └── (* values)    values tuple (can be understand when relation is done with (m/v/i/*sorted_key))
│   └── v                     vertex
│       └── <id>              vertex index id
│           └── (* values)    values tuple (can be understand when relation is done with (m/v/i/*sorted_key))
├── e                         EDGE (DATA & RELATIONS)
│   └── <id>                  edge id
│       ├── d                 edge data
│       ├── i                 edge indexes containing this edge (this is a list of index ids)
│       ├── iv                incoming vertices
│       │   └── c             incoming vertices count
│       ├── ov                outgoing vertices
│       │   └── c             outgoing vertices count
│       └── bv                both vertices
│           └── c             both vertices count
└── v                         VERTICES (DATA & RELATIONS)
    └── <id>                  vertex id
        ├── d                 vertex data
        ├── i                 vertex indexes containing this vertex (this is a list of index ids)
        ├── ie                incoming edges
        │   └── c             incoming edges count
        ├── iv                incoming vertices
        │   └── c             incoming vertices count
        ├── oe                outgoing edges
        │   └── c             outgoing edges count
        ├── ov                outgoing vertices
        │   └── c             outgoing vertices count
        ├── be                both edges
        │   └── c             both edges count
        └── bv                both vertices
            └── c             both vertices count

"""

import re

CHUNK_SIZE                          = 1000

PREPARED                            = 'prep'
FORBIDDEN_KEY                       = 'xxxx'

METADATA_EDGE_COUNTER               = 'm/e/c'
METADATA_EDGE_REMOVED_COUNTER       = 'm/e/r'
METADATA_EDGE_ID_LIST_PREFIX        = 'm/e/id'
METADATA_EDGE_INDEX_COUNTER         = 'm/e/ci'
METADATA_EDGE_INDEX_LIST            = 'm/e/il'
METADATA_EDGE_INDEX_FIELDS_PREFIX   = 'm/e/if'
METADATA_EDGE_INDEX_PREFIX          = 'm/e/i'
METADATA_VERTEX_COUNTER             = 'm/v/c'
METADATA_VERTEX_REMOVED_COUNTER     = 'm/v/r'
METADATA_VERTEX_ID_LIST_PREFIX      = 'm/v/id'
METADATA_VERTEX_INDEX_COUNTER       = 'm/v/ci'
METADATA_VERTEX_INDEX_LIST          = 'm/v/il'
METADATA_VERTEX_INDEX_FIELDS_PREFIX = 'm/v/if'
METADATA_VERTEX_INDEX_PREFIX        = 'm/v/i'

DATA_SUFFIX                         = 'd'
IN_EDGES_SUFFIX                     = 'ie'
IN_VERTICES_SUFFIX                  = 'iv'
OUT_EDGES_SUFFIX                    = 'oe'
OUT_VERTICES_SUFFIX                 = 'ov'
BOTH_EDGES_SUFFIX                   = 'be'
BOTH_VERTICES_SUFFIX                = 'bv'
COUNT_SUFFIX                        = 'c'
COUNT_REMOVE_SUFFIX                 = 'r'

KIND_EDGE                           = 'e'
KIND_VERTEX                         = 'v'
KIND_INDEX                          = 'i'

DATA_REGEXP_NOGROUP                 = '^%s/[0-9]+/%s$'
DATA_REGEXP_GROUP                   = '^%s/([0-9]+)/%s$'
EDGE_DATA_REGEXP                    = DATA_REGEXP_NOGROUP % (KIND_EDGE, DATA_SUFFIX)
VERTEX_DATA_REGEXP                  = DATA_REGEXP_NOGROUP % (KIND_VERTEX, DATA_SUFFIX)
EDGE_DATA_PATTERN                   = re.compile(DATA_REGEXP_GROUP % (KIND_EDGE, DATA_SUFFIX))
VERTEX_DATA_PATTERN                 = re.compile(DATA_REGEXP_GROUP % (KIND_VERTEX, DATA_SUFFIX))


def build_key(*args):
    return '/'.join([str(arg) for arg in args])



# -*- coding:utf-8 -*-

import random
from collections import defaultdict
from functools import reduce
from itertools import chain, tee

from grapheekdb.lib.undef import UNDEFINED
from grapheekdb.lib.readwrite import GraphReadWrite
from grapheekdb.lib.nx import GraphNx
from grapheekdb.lib.validations import check_valid_data
from grapheekdb.lib.exceptions import GrapheekDataException, GrapheekDataPreparationFailedException, GrapheekIndexAlreadyExistsException
from grapheekdb.lib.exceptions import GrapheekIndexCreationFailedException, GrapheekIndexRemovalFailedException, GrapheekNoSuchTraversalException
from grapheekdb.lib.exceptions import GrapheekMissingKeyException, GrapheekInvalidDataTypeException
from grapheekdb.lib.exceptions import GrapheekUnknownScriptException, GrapheekUnknownAlias
from grapheekdb.lib.exceptions import GrapheekUnknownMethod, GrapheekMixedKindException
from grapheekdb.lib.expressions import eval_expr

from grapheekdb.backends.data.keys import build_key, PREPARED
from grapheekdb.backends.data.keys import METADATA_EDGE_COUNTER, METADATA_EDGE_INDEX_COUNTER, METADATA_EDGE_INDEX_LIST, METADATA_EDGE_INDEX_FIELDS_PREFIX, METADATA_EDGE_INDEX_PREFIX
from grapheekdb.backends.data.keys import METADATA_VERTEX_COUNTER, METADATA_VERTEX_INDEX_COUNTER, METADATA_VERTEX_INDEX_LIST, METADATA_VERTEX_INDEX_FIELDS_PREFIX, METADATA_VERTEX_INDEX_PREFIX
from grapheekdb.backends.data.keys import METADATA_VERTEX_REMOVED_COUNTER, METADATA_EDGE_REMOVED_COUNTER
from grapheekdb.backends.data.keys import DATA_SUFFIX, IN_EDGES_SUFFIX, IN_VERTICES_SUFFIX, OUT_EDGES_SUFFIX, OUT_VERTICES_SUFFIX, BOTH_EDGES_SUFFIX, BOTH_VERTICES_SUFFIX
from grapheekdb.backends.data.keys import COUNT_SUFFIX, KIND_VERTEX, KIND_EDGE, KIND_INDEX
from grapheekdb.backends.data.keys import METADATA_EDGE_ID_LIST_PREFIX, METADATA_VERTEX_ID_LIST_PREFIX, CHUNK_SIZE
from grapheekdb.backends.data.filtertools import build_filter_funcs, filter_entities
from grapheekdb.backends.data.indexes import ExactIndex, normalize_value
from grapheekdb.backends.data.operations import Addition, Removal
from grapheekdb.backends.data.optimizer import Optimizer


# misc iterators :
def _jump_iterator(_graph, _kind, _entity_id_iterator, _traversal, _neighbors_cache, **filters):
    for entity_id in _entity_id_iterator:
        for neighbor_id in iter(_graph._neighbors(_kind, entity_id, _traversal, _neighbors_cache, False, **filters)):
            yield neighbor_id


def _random_iterator(_graph, _kind, _entity_id_iterator, _traversal, _neighbors_cache, **filters):
    for entity_id in _entity_id_iterator:
        for neighbor_id in iter(_graph._neighbors(_kind, entity_id, _traversal, _neighbors_cache, True, **filters)):
            yield neighbor_id


def check_base_node(entity):
    if not(isinstance(entity, Node)):
        raise GrapheekInvalidDataTypeException('%s is not a node' % (entity,))


def check_and_get_count(args):
    count_args = len(args)
    assert(count_args < 2)
    # how many recursive jumps ? :
    count = 1
    if count_args == 1:
        count = int(args[0])
        assert(count >= 0)
    return count


class DotExport(object):
    """
    This class is mainly intended to be used with an IPython Graphviz Magics extension
    You can start IPython by launching in a shell
    ipython notebook

    Then create a new notebook and install extension :
    %install_ext https://raw.github.com/cjdrake/ipython-magic/master/gvmagic.py

    Load extension :
    %load_ext gvmagic

    Then you can see an entity iterator by doing, for instance :
    %dotobj g.V(...).dot()

    The EntityIterator.dot method instantiates a DotExport which itself has a .to_dot method as expected by gvmagic...
    """

    def __init__(self, entity_iterator, node_label=None, edge_label=None, limit=100):
        self._entity_iterator = entity_iterator
        self._graph = entity_iterator._graph
        self.node_label = node_label
        self.edge_label = edge_label
        self.limit = int(limit)
        assert(self.limit > 0)

    def __str__(self):
        return self.to_dot()

    def _clean_label(self, label):
        import re
        s = re.sub('[^0-9a-zA-Z_\s]', '', str(label))  # Needs sth less intrusive
        return s

    def _node_label(self, node_id):
        node = Node(node_id, self._graph)
        label = str(node_id)  # Default value
        if self.node_label is not None:
            label = node.data().get(self.node_label, label)
        return self._clean_label(label)

    def _edge_struct(self, edge_id):
        edge = Edge(edge_id, self._graph)
        label = ''  # Default value
        if self.edge_label is not None:
            label = edge.data().get(self.edge_label, label)
        # get src and tgt node ids :
        src_id = edge._src_id()
        tgt_id = edge._tgt_id()
        return (self._clean_label(label), src_id, tgt_id)

    def to_dot(self):
        entity_iterator = self._entity_iterator
        kind = entity_iterator._src_kind
        assert(kind in [KIND_VERTEX, KIND_EDGE])
        node_collections = []
        edge_collections = []
        if kind == KIND_EDGE:
            # Collecting label, ids and related node ids :
            edge_node_ids = set()
            for edge_id in set(entity_iterator.limit(self.limit)._iterate(on_item=False)):
                label, src_id, tgt_id = self._edge_struct(edge_id)
                edge_collections.append((label, src_id, tgt_id))
                edge_node_ids.add(src_id)
                edge_node_ids.add(tgt_id)
            # Collecting nodes :
            for node_id in edge_node_ids:
                label = self._node_label(node_id)
                node_collections.append((node_id, label))
        else:  # KIND_VERTEX
            node_edge_ids = set()
            node_ids = set(entity_iterator.limit(self.limit)._iterate(on_item=False))
            # Collecting label, ids and related edge ids :
            for node_id in node_ids:
                label = self._node_label(node_id)
                node_collections.append((node_id, label))
                for edge_id in self._graph._neighbors(KIND_VERTEX, node_id, BOTH_EDGES_SUFFIX):
                    if edge_id not in node_edge_ids:
                        edge_label, edge_src_id, edge_tgt_id = self._edge_struct(edge_id)
                        if (edge_src_id in node_ids) and (edge_tgt_id in node_ids):
                            edge_collections.append((edge_label, edge_src_id, edge_tgt_id))
                            node_edge_ids.add(edge_id)
        # Finally build the dot string :
        lines = []
        lines.append('digraph G {')
        for node_id, label in node_collections:
            lines.append('"n%s" [' % node_id)
            lines.append('label ="%s"' % label)
            lines.append('];')
        for label, src_id, tgt_id in edge_collections:
            lines.append('"n%s" -> "n%s" [label="%s"]' % (src_id, tgt_id, label))
        lines.append('}')
        return '\n'.join(lines)


class EntityAggregate(object):

    def __init__(self, graph, entity_kind, entity_iterator, pair_iterator=None, agg='+', expr=None, asc=False):
        self._graph = graph
        self._entity_kind = entity_kind
        self._entity_iterator = entity_iterator
        self._pair_iterator = pair_iterator
        assert(agg in (None, '+', '%'))
        self._agg = agg
        self._expr = expr
        self._mult = 1 if asc else -1
        # -------------------------------
        self._current_iteration = None

    def _get_results(self):
        # Special case : if self._agg is None, then it means that current EntityAggregate
        # has been created by another EntityAggregate (for instance, by method limit)
        # so we'll just iterate on _pair_iterator
        if self._agg is None:
            return list(self._pair_iterator)
        if self._agg == '+':
            counters = defaultdict(int)
        else:
            counters = defaultdict(float)
        total = 0
        entity_data_cache = {}
        for entity_id in self._entity_iterator._iterate(on_item=False):
            if self._expr is None:
                value = 1
            else:
                # Safe eval of expression :
                entity_data = entity_data_cache.get(entity_id, None)
                if entity_data is None:
                    entity_data = self._graph._item_from_id(self._entity_kind, entity_id).data()
                    entity_data_cache[entity_id] = entity_data
                context = entity_data
                context['_'] = self._entity_iterator._context
                try:
                    value = eval_expr(context, self._expr)
                finally:
                    del context['_']
            counters[entity_id] += value
            total += value
        if total and (self._agg == '%'):
            for entity_id, value in list(counters.items()):
                counters[entity_id] = float(counters[entity_id]) / float(total)  # This may raise an exception, it's up to the user to handle it...
        ordered_counters = sorted(list(counters.items()), key=lambda t: t[1] * self._mult)
        results = []
        for entity_id, count in ordered_counters:
            entity = self._graph._item_from_id(self._entity_kind, entity_id)
            results.append((entity, count))
        return results

    def __iter__(self):
        return self

    def next(self):  # pragma : no cover
        # Python 2 compatibility
        return self.__next__()

    def __next__(self):
        if self._current_iteration is None:
            self._current_iteration = iter(self._get_results())
        try:
            return next(self._current_iteration)
        except StopIteration:
            self._current_iteration = None
        raise StopIteration

    def entities(self):
        def entity_iterator(iterator):
            for entity, _ in iterator:
                yield entity.get_id()
        return EntityIterator(self._graph, self._entity_kind, {}, entity_iterator(self), parent=self._entity_iterator)

    def limit(self, number):
        def limit_iterator(iterator, number):
            counter = 0
            for entity_id, count in iterator:
                if counter < number:
                    yield (entity_id, count)
                else:
                    break  # This line is REALLY important (it stops outer iteration)
                counter += 1
        return EntityAggregate(self._graph, self._entity_kind, self._entity_iterator, limit_iterator(self, number), agg=None, asc=None)


class EntityIterator(object):

    def __init__(self, graph, src_kind, filters, id_iterator, parent=None, neighbors_cache=None, context=None):
        assert(id_iterator is not None)
        self._graph = graph
        self._src_kind = src_kind
        self._filters = filters
        self._filter_funcs = build_filter_funcs(**filters)
        self._initial_id_iterator, self._id_iterator = tee(id_iterator)
        self._parent = parent
        self._neighbors_cache = neighbors_cache if neighbors_cache is not None else {}
        self._release_neighbors_cache = neighbors_cache is None
        if context is None:
            self._context = {} if parent is None else parent._context
        else:
            self._context = context
        self._current_iteration = None

    def __next__(self):
        if self._current_iteration is None:
            self._current_iteration = self._clone_initial()._iterate(on_item=True)
        try:
            return next(self._current_iteration)
        except StopIteration:
            self._current_iteration = None
        raise StopIteration

    def __iter__(self):
        return self

    def __contains__(self, entity):
        try:
            if entity._kind != self._src_kind:
                return False
        except:
            return False
        entity_id = entity.get_id()
        for it_entity_id in self._clone_initial()._iterate(on_item=False):
            if it_entity_id == entity_id:
                return True
        return False

    def __le__(self, it2):
        return self._graph.issubset(self, it2)

    def __ge__(self, it2):
        return self._graph.issuperset(self, it2)

    def __or__(self, it2):
        # | => union
        return self._graph.union(self, it2)

    def __and__(self, it2):
        # & => intersection
        return self._graph.intersection(self, it2)

    def __sub__(self, it2):
        # - => difference
        return self._graph.difference(self, it2)

    def __xor__(self, it2):
        # ^ => symmetric_difference
        return self._graph.symmetric_difference(self, it2)

    def _finalize(self):
        if self._release_neighbors_cache:
            # Implicit cache removal
            self._neighbors_cache = {}

    def _iterate(self, on_item=False):
        for item in filter_entities(self._graph, self._src_kind, self._id_iterator, self._filter_funcs, on_item=on_item):
            yield item
        self._finalize()

    def _clone_initial(self):
        self._initial_id_iterator, id_iterator = tee(self._initial_id_iterator)
        return EntityIterator(self._graph, self._src_kind, self._filters, id_iterator, self._parent)

    def _jump(self, _kind, _traversal, _random, *args, **filters):
        jumping_iterator = _random_iterator if _random else _jump_iterator
        # TODO : Refactor this function : it can certainly be rewritten in a far easier way
        count = check_and_get_count(args)
        current_iterator = self
        parent_kind = self._src_kind
        current_kind = _kind
        current_parent = self
        # (potentially) recursive jumping
        #_jump_iterator(graph, kind, entity_id_iterator, traversal, neighbors_cache, filters)
        for _ in range(count):
            current_iterator = EntityIterator(
                self._graph,
                current_kind,
                filters,
                jumping_iterator(
                    self._graph,
                    parent_kind,
                    current_iterator._iterate(),
                    _traversal,
                    self._neighbors_cache,
                    **filters
                ),
                parent=current_parent
            )
            parent_kind = current_kind
            current_kind = KIND_VERTEX
            current_parent = current_iterator
        return current_iterator

    def next(self):
        # Python 2 compatibility
        return self.__next__()

    def inV(self, *args, **filters):
        return self._jump(KIND_VERTEX, IN_VERTICES_SUFFIX, False, *args, **filters)

    def outV(self, *args, **filters):
        return self._jump(KIND_VERTEX, OUT_VERTICES_SUFFIX, False, *args, **filters)

    def bothV(self, *args, **filters):
        return self._jump(KIND_VERTEX, BOTH_VERTICES_SUFFIX, False, *args, **filters)

    def in_(self, *args, **filters):
        assert(self._src_kind == KIND_VERTEX)
        count = check_and_get_count(args)
        if count == 1:
            return self.inE(**filters).inV()
        return self.in_(count - 1, **filters).inE(*args, **filters).inV()

    def out_(self, *args, **filters):
        assert(self._src_kind == KIND_VERTEX)
        count = check_and_get_count(args)
        if count == 1:
            return self.outE(**filters).outV()
        return self.out_(count - 1, **filters).outE(*args, **filters).outV()

    def both_(self, *args, **filters):
        assert(self._src_kind == KIND_VERTEX)
        count = check_and_get_count(args)

        def neighbor_iterator(graph, it, filters, neighbors_cache, filter_funcs):
            entity_ids = list(it)
            for entity_id in entity_ids:
                neighbor_edge_id_iterator = graph._neighbors(KIND_VERTEX, entity_id, OUT_EDGES_SUFFIX, neighbors_cache, **filters)
                for neighbor_edge_id in filter_entities(self._graph, KIND_EDGE, neighbor_edge_id_iterator, filter_funcs):
                    for neighbor_node_id in graph._neighbors(KIND_EDGE, neighbor_edge_id, OUT_VERTICES_SUFFIX, neighbors_cache):
                        if int(entity_id) != int(neighbor_node_id):
                            yield neighbor_node_id
            for entity_id in entity_ids:
                neighbor_edge_id_iterator = graph._neighbors(KIND_VERTEX, entity_id, IN_EDGES_SUFFIX, neighbors_cache, **filters)
                for neighbor_edge_id in filter_entities(self._graph, KIND_EDGE, neighbor_edge_id_iterator, filter_funcs):
                    for neighbor_node_id in graph._neighbors(KIND_EDGE, neighbor_edge_id, IN_VERTICES_SUFFIX, neighbors_cache):
                        if int(entity_id) != int(neighbor_node_id):
                            yield neighbor_node_id

        filter_funcs = build_filter_funcs(**filters)
        if count == 1:
            return EntityIterator(
                self._graph,
                KIND_VERTEX,
                {},
                neighbor_iterator(self._graph, self._iterate(), filters, self._neighbors_cache, filter_funcs),
                parent=self,
                neighbors_cache=self._neighbors_cache
            )
        entity_iterator = self.both_(count - 1, **filters)
        return EntityIterator(
            self._graph,
            KIND_VERTEX,
            {},
            neighbor_iterator(self._graph, entity_iterator._iterate(), filters, self._neighbors_cache, filter_funcs),
            parent=entity_iterator,
            neighbors_cache=self._neighbors_cache
        )

    # NOTE : for xxE methods, below, there's no *args :
    # currently *args is aimed to contain only number of jumps
    # so only 1st item of args is used
    # if args is empty, only one jump is done
    # So this is "recursive" jumping
    # But chaining xxE methods wouldn't make sense as Edge outer entities are always Nodes

    def inE(self, *args, **filters):
        if self._src_kind == KIND_EDGE:
            raise GrapheekNoSuchTraversalException
        return self._jump(KIND_EDGE, IN_EDGES_SUFFIX, False, **filters)  # no *args for inE ()

    def outE(self, *args, **filters):
        if self._src_kind == KIND_EDGE:
            raise GrapheekNoSuchTraversalException
        return self._jump(KIND_EDGE, OUT_EDGES_SUFFIX, False, **filters)

    def bothE(self, *args, **filters):
        if self._src_kind == KIND_EDGE:
            raise GrapheekNoSuchTraversalException
        return self._jump(KIND_EDGE, BOTH_EDGES_SUFFIX, False, **filters)

    def random(self, _method, *args, **filters):
        if self._src_kind == KIND_VERTEX:
            try:
                kind, traversal = {
                    'inV': (KIND_VERTEX, IN_VERTICES_SUFFIX),
                    'outV': (KIND_VERTEX, OUT_VERTICES_SUFFIX),
                    'bothV': (KIND_VERTEX, BOTH_VERTICES_SUFFIX),
                    'inE': (KIND_EDGE, IN_EDGES_SUFFIX),
                    'outE': (KIND_EDGE, OUT_EDGES_SUFFIX),
                    'bothE': (KIND_EDGE, BOTH_EDGES_SUFFIX)
                }[_method]
            except KeyError:
                raise GrapheekUnknownMethod(_method)
        elif self._src_kind == KIND_EDGE:
            try:
                kind, traversal = {
                    'inV': (KIND_VERTEX, IN_VERTICES_SUFFIX),
                    'outV': (KIND_VERTEX, OUT_VERTICES_SUFFIX),
                    'bothV': (KIND_VERTEX, BOTH_VERTICES_SUFFIX)
                }[_method]
            except KeyError:
                raise GrapheekUnknownMethod(_method)
        return self._jump(KIND_EDGE, BOTH_EDGES_SUFFIX, True, **filters)

    def dedup(self, *aliases):
        def dedup_iterator(entity_iterator, aliases):
            if aliases:
                alias_known = defaultdict(set)
            else:
                known = set()
            for entity_id in entity_iterator._iterate():
                if aliases:
                    keys = []
                    for alias in aliases:
                        ctx_entity = entity_iterator._context.get(alias, None)
                        if ctx_entity is None:
                            raise GrapheekUnknownAlias("'%s' alias is not defined" % (alias,))
                        keys.append((ctx_entity._kind, ctx_entity.get_id()))
                    known = alias_known[tuple(keys)]
                if entity_id in known:
                    continue
                known.add(entity_id)
                yield entity_id
        return EntityIterator(self._graph, self._src_kind, {}, dedup_iterator(self, aliases), parent=self)

    def without(self, *aliases):
        def without_iterator(entity_iterator, aliases):
            clone_entity_iterator = entity_iterator
            for entity_id in clone_entity_iterator._iterate(on_item=False):  # Iterating updates the context ...
                member = False
                for alias in aliases:
                    excluded_entity = entity_iterator._context.get(alias, UNDEFINED)
                    if excluded_entity._kind == self._src_kind and excluded_entity.get_id() == entity_id:
                        member = True
                        break
                if not member:
                    yield entity_id
        return EntityIterator(self._graph, self._src_kind, {}, without_iterator(self, aliases), parent=self)

    def idata(self, *args, **kwargs):
        for item in self:
            yield item.data(*args, **kwargs)

    def data(self, *args, **kwargs):
        return list(self.idata(*args, **kwargs))

    def limit(self, count):
        def limit_iterator(entity_id_iterator, count):
            counter = 0
            for entity_id in entity_id_iterator:
                if counter < count:
                    yield entity_id
                else:
                    break  # This line is REALLY important (it stops outer iteration)
                counter += 1
        return EntityIterator(self._graph, self._src_kind, {}, limit_iterator(self._iterate(), count), parent=self)

    def aka(self, alias):
        def alias_iterator(entity_id_iterator, alias):
            graph = self._graph
            for entity_id in entity_id_iterator:
                self._context[alias] = graph._item_from_id(self._src_kind, entity_id)
                yield entity_id
        return EntityIterator(self._graph, self._src_kind, {}, alias_iterator(self._iterate(), alias), parent=self)

    # Following methods don't return iterator :

    def count(self):
        return sum(1 for _ in self._iterate())  # space efficient counter

    def remove(self):
        remove_func = self._graph._remove_node if self._src_kind == KIND_VERTEX else self._graph._remove_edge
        for entity_id in self._iterate():
            remove_func(entity_id)

    def update(self, **updates):
        # Do the update (not done while iterating because *every* entity must be updated before starting to yield)
        txn = self._graph._transaction_begin()
        try:
            for entity_id in self._iterate(on_item=False):
                self._graph._bulk_update_data(txn, self._src_kind, entity_id, **updates)
            self._graph._transaction_commit(txn)
        except Exception as e:
            self._graph._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def all(self):
        return list(self)

    def ids(self):
        return list(self._iterate(on_item=False))

    def collect(self, *aliases):
        result = []
        for entity in self:  # Iterating updates the context ...
            result.append([self._context.get(alias, None) for alias in aliases])
        return result

    def sum(self, expr=None, asc=False):
        return EntityAggregate(self._graph, self._src_kind, self, pair_iterator=None, agg='+', expr=expr, asc=asc)

    def percent(self, expr=None, asc=False):
        return EntityAggregate(self._graph, self._src_kind, self, pair_iterator=None, agg='%', expr=expr, asc=asc)

    def _context_iterator(self):
        for entity_id in self._iterate():  # Iterating updates the context ...
            yield self._context.copy()

    def call(self, _fname, *aliases, **params):
        # Same as request but ignore return
        self.request(_fname, *aliases, **params)

    def request(self, _fname, *aliases, **params):
        fn = self._graph._get_server_script(_fname)
        return fn(self._graph, self._context_iterator(), *aliases, **params)

    def dot(self, node_label=None, edge_label=None, limit=100):
        return DotExport(self, node_label, edge_label, limit)

    def _dot_str(self, node_label=None, edge_label=None, limit=100):  # pragma : no cover
        return str(DotExport(self, node_label, edge_label, limit))


class Entity(object):

    def __init__(self, entity_id, graph):
        self._entity_id = int(entity_id)  # Depending on backend, sometimes, id are stored as string
        self._graph = graph

    def __repr__(self):
        kind = 'node' if self._kind == KIND_VERTEX else 'edge'
        return '<%s id:%s data:%s>' % (kind, self._entity_id, repr(self.data()))

    def __eq__(self, other):
        return other.__class__ == self.__class__ and other.get_id() == self.get_id() and other._graph == self._graph

    def __ne__(self, other):
        # Cf Python Doc : http://docs.python.org/2/reference/datamodel.html#object.__ne__
        # "When defining __eq__(), one should also define __ne__() so that the operators will behave as expected"
        return not self.__eq__(other)

    def __getattr__(self, attr):
        try:
            return self._graph._get_data(None, self._kind, self._entity_id)[attr]
        except KeyError:
            raise AttributeError("%r instance has no attribute %r" % (self.__class__, attr))

    def __setattr__(self, attr, value):
        if attr.startswith('_'):
            self.__dict__.update({attr: value})
        else:
            self._graph._update_data(self._kind, self._entity_id, attr, value)

    def _neighbors(self, _traversal, _random, **filters):
        return self._graph._neighbors(self._kind, self._entity_id, _traversal, _cache=None, _random=_random, **filters)

    def _jump(self, _tgt_kind, _traversal, _random, *args, **filters):
        jumping_iterator = _random_iterator if _random else _jump_iterator
        # TODO : Factor code with EntityIterator._jump
        count_args = len(args)
        assert(count_args < 2)
        # how many recursive jumps ? :
        count = 1
        if count_args == 1:
            count = args[0]
        current_iterator = self._neighbors(_traversal, _random, **filters)
        parent_kind = self._kind
        current_kind = _tgt_kind
        current_parent = None
        # (potentially) recursive jumping
        for i in range(count):
            IT = current_iterator if i == 0 else jumping_iterator(self._graph, parent_kind, current_iterator._iterate(), _traversal, {}, **filters)
            current_iterator = EntityIterator(
                self._graph,
                current_kind,
                filters,
                IT,
                parent=current_parent
            )
            parent_kind = current_kind
            current_kind = KIND_VERTEX
            current_parent = current_iterator
        return current_iterator

    def get(self, field, default=None):
        try:
            return self.__getattr__(field)
        except AttributeError:
            return default

    def inV(self, *args, **filters):
        return self._jump(KIND_VERTEX, IN_VERTICES_SUFFIX, False, *args, **filters)

    def outV(self, *args, **filters):
        return self._jump(KIND_VERTEX, OUT_VERTICES_SUFFIX, False, *args, **filters)

    def bothV(self, *args, **filters):
        return self._jump(KIND_VERTEX, BOTH_VERTICES_SUFFIX, False, *args, **filters)

    def random(self, _method, **filters):
        try:
            kind, traversal = {
                'inV': (KIND_VERTEX, IN_VERTICES_SUFFIX),
                'outV': (KIND_VERTEX, OUT_VERTICES_SUFFIX),
                'bothV': (KIND_VERTEX, BOTH_VERTICES_SUFFIX)
            }[_method]
        except KeyError:
            raise GrapheekUnknownMethod(_method)
        return self._jump(kind, traversal, True, **filters)

    def data(self, *args, **kwargs):
        return self._graph._get_data(None, self._kind, self._entity_id, *args, **kwargs)

    def get_id(self):
        return self._entity_id

    def update(self, **updates):
        self._graph._bulk_update_data(None, self._kind, self._entity_id, **updates)

    def aka(self, alias):
        def self_iterator(entity_id):
            yield entity_id
        return EntityIterator(self._graph, self._kind, {}, self_iterator(self.get_id()), context={alias: self})


class Node(Entity):

    def __init__(self, entity_id, graph):
        super(Node, self).__init__(entity_id, graph)
        self._kind = KIND_VERTEX

    def __change_denorm_data(self, lst_method, counter_method, *counter_args):
        root_key = KIND_VERTEX + '/' + str(self._entity_id) + '/'
        # traversal denormalized lists
        lst_method(root_key + IN_EDGES_SUFFIX)
        lst_method(root_key + IN_VERTICES_SUFFIX)
        lst_method(root_key + OUT_EDGES_SUFFIX)
        lst_method(root_key + OUT_VERTICES_SUFFIX)
        lst_method(root_key + BOTH_EDGES_SUFFIX)
        lst_method(root_key + BOTH_VERTICES_SUFFIX)
        # traversal denormalized counters
        counter_method(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
        counter_method(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
        counter_method(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
        counter_method(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
        counter_method(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)
        counter_method(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX, *counter_args)

    def _add_denorm_data(self):
        operation = Addition()
        self.__change_denorm_data(operation.init_lst, operation.set, 0)
        return operation

    def _remove_denorm_data(self):
        operation = Removal()
        self.__change_denorm_data(operation.remove, operation.remove)
        return operation

    def _add_denorm_src_data(self, target, edge):
        source_id = self._entity_id
        target_id = target.get_id()
        edge_id = edge.get_id()
        root_key = KIND_VERTEX + '/' + str(source_id) + '/'
        s_edge_id = str(edge_id)
        s_target_id = str(target_id)
        # building operation :
        operation = Addition()
        # lists
        operation.append_to_lst(root_key + OUT_EDGES_SUFFIX, edge_id)
        operation.append_to_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
        operation.append_to_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
        # counters
        operation.update_inc(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        # create a key/value for links between source and target (aimed to be used with indexes)
        operation.set(root_key + OUT_EDGES_SUFFIX + '/' + s_edge_id, 1)
        operation.set(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id, 1)
        operation.set(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id, 1)
        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id, 1)
        # ---
        return operation

    def _remove_denorm_src_data(self, target, edge):
        source_id = self._entity_id
        target_id = target.get_id()
        edge_id = edge.get_id()
        root_key = KIND_VERTEX + '/' + str(source_id) + '/'
        s_edge_id = str(edge_id)
        s_target_id = str(target_id)
        # building operation :
        operation = Removal()
        # lists
        operation.remove_from_lst(root_key + OUT_EDGES_SUFFIX, edge_id)
        operation.remove_from_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
        operation.remove_from_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
        operation.remove_from_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
        # counters
        operation.update_dec(root_key + OUT_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + OUT_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        # create a key/value for links between source and target (aimed to be used with indexes)
        operation.remove(root_key + OUT_EDGES_SUFFIX + '/' + s_edge_id)
        operation.remove(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id)
        operation.remove(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id)
        # ---
        return operation

    def _add_denorm_tgt_data(self, source, edge):
        target_id = self._entity_id
        source_id = source.get_id()
        edge_id = edge.get_id()
        root_key = KIND_VERTEX + '/' + str(target_id) + '/'
        s_edge_id = str(edge_id)
        s_source_id = str(source_id)
        # building operation :
        operation = Addition()
        # lists
        operation.append_to_lst(root_key + IN_EDGES_SUFFIX, edge_id)
        operation.append_to_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
        operation.append_to_lst(root_key + IN_VERTICES_SUFFIX, source_id)
        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
        # counters
        operation.update_inc(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_inc(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        # create a key/value for links between source and target (aimed to be used with indexes)
        operation.set(root_key + IN_EDGES_SUFFIX + '/' + s_edge_id, 1)
        operation.set(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id, 1)
        operation.set(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id, 1)
        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id, 1)
        # ---
        return operation

    def _remove_denorm_tgt_data(self, source, edge):
        target_id = self._entity_id
        source_id = source.get_id()
        edge_id = edge.get_id()
        root_key = KIND_VERTEX + '/' + str(target_id) + '/'
        s_edge_id = str(edge_id)
        s_source_id = str(source_id)
        # building operation :
        operation = Removal()
        # lists
        operation.remove_from_lst(root_key + IN_EDGES_SUFFIX, edge_id)
        operation.remove_from_lst(root_key + BOTH_EDGES_SUFFIX, edge_id)
        operation.remove_from_lst(root_key + IN_VERTICES_SUFFIX, source_id)
        operation.remove_from_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
        # counters
        operation.update_dec(root_key + IN_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + BOTH_EDGES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + IN_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        operation.update_dec(root_key + BOTH_VERTICES_SUFFIX + '/' + COUNT_SUFFIX)
        # create a key/value for links between source and target (aimed to be used with indexes)
        operation.remove(root_key + IN_EDGES_SUFFIX + '/' + s_edge_id)
        operation.remove(root_key + BOTH_EDGES_SUFFIX + '/' + s_edge_id)
        operation.remove(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id)
        # ---
        return operation

    def in_(self, *args, **filters):
        def self_iterator(entity_id):
            yield entity_id
        entity_iterator = EntityIterator(self._graph, KIND_VERTEX, {}, self_iterator(self.get_id()), context={})
        return entity_iterator.in_(*args, **filters)

    def out_(self, *args, **filters):
        def self_iterator(entity_id):
            yield entity_id
        entity_iterator = EntityIterator(self._graph, KIND_VERTEX, {}, self_iterator(self.get_id()), context={})
        return entity_iterator.out_(*args, **filters)

    def both_(self, *args, **filters):
        def self_iterator(entity_id):
            yield entity_id
        entity_iterator = EntityIterator(self._graph, KIND_VERTEX, {}, self_iterator(self.get_id()), context={})
        return entity_iterator.both_(*args, **filters)

    def inE(self, **filters):
        return self._jump(KIND_EDGE, IN_EDGES_SUFFIX, False, **filters)

    def outE(self, **filters):
        return self._jump(KIND_EDGE, OUT_EDGES_SUFFIX, False, **filters)

    def bothE(self, **filters):
        return self._jump(KIND_EDGE, BOTH_EDGES_SUFFIX, False, **filters)

    def random(self, _method, **filters):
        try:
            kind, traversal = {
                'inV': (KIND_VERTEX, IN_VERTICES_SUFFIX),
                'outV': (KIND_VERTEX, OUT_VERTICES_SUFFIX),
                'bothV': (KIND_VERTEX, BOTH_VERTICES_SUFFIX),
                'inE': (KIND_EDGE, IN_EDGES_SUFFIX),
                'outE': (KIND_EDGE, OUT_EDGES_SUFFIX),
                'bothE': (KIND_EDGE, BOTH_EDGES_SUFFIX)
            }[_method]
        except KeyError:
            raise GrapheekUnknownMethod(_method)
        return self._jump(kind, traversal, True, **filters)

    def remove(self):
        self._graph._remove_node(self._entity_id)


class Edge(Entity):

    def __init__(self, entity_id, graph):
        super(Edge, self).__init__(entity_id, graph)
        self._kind = KIND_EDGE

    def _src_id(self):
        return self._graph._get_lst(None, build_key(KIND_EDGE, self._entity_id, IN_VERTICES_SUFFIX))[0]

    def _tgt_id(self):
        return self._graph._get_lst(None, build_key(KIND_EDGE, self._entity_id, OUT_VERTICES_SUFFIX))[0]

    def _add_denorm_data(self, source, target):
        edge_id = self._entity_id
        source_id = source.get_id()
        target_id = target.get_id()
        root_key = KIND_EDGE + '/' + str(edge_id) + '/'
        s_source_id = str(source_id)
        s_target_id = str(target_id)
        # building operation :
        operation = Addition()
        # traversal denormalized lists
        operation.append_to_lst(root_key + IN_VERTICES_SUFFIX, source_id)
        operation.append_to_lst(root_key + OUT_VERTICES_SUFFIX, target_id)
        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, source_id)
        operation.append_to_lst(root_key + BOTH_VERTICES_SUFFIX, target_id)
        # create a key/value for links between source, target and current edge (aimed to be used with indexes)
        operation.set(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id, 1)
        operation.set(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id, 1)
        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id, 1)
        operation.set(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id, 1)
        # ---
        return operation

    def _remove_denorm_data(self, source, target):
        edge_id = self._entity_id
        source_id = source.get_id()
        target_id = target.get_id()
        root_key = KIND_EDGE + '/' + str(edge_id) + '/'
        s_source_id = str(source_id)
        s_target_id = str(target_id)
        # building operation
        operation = Removal()
        # traversal denormalized lists
        operation.remove(root_key + IN_VERTICES_SUFFIX)
        operation.remove(root_key + OUT_VERTICES_SUFFIX)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX)
        # remove key/value for links between source, target and current edge (aimed to be used with indexes)
        operation.remove(root_key + IN_VERTICES_SUFFIX + '/' + s_source_id)
        operation.remove(root_key + OUT_VERTICES_SUFFIX + '/' + s_target_id)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_source_id)
        operation.remove(root_key + BOTH_VERTICES_SUFFIX + '/' + s_target_id)
        # ---
        return operation

    def remove(self):
        self._graph._remove_edge(self._entity_id)


class BaseGraph(GraphReadWrite, GraphNx):

    def close(self):
        if not self._closed:
            self._db_close()
            self._closed = True

    def __initialize__(self):
        # "Rebuilding" indexes :
        txn = self._transaction_begin()
        try:
            # Get vertex and edge count (useful for queries where we need to compare index perf and seq scan) :
            self._node_count = self._get(txn, METADATA_VERTEX_COUNTER) - self._get(txn, METADATA_VERTEX_REMOVED_COUNTER)
            self._edge_count = self._get(txn, METADATA_EDGE_COUNTER) - self._get(txn, METADATA_EDGE_REMOVED_COUNTER)
            # Instantiating vertex and edge indexes
            self._node_indexes = []
            self._edge_indexes = []
            #   1st vertex indexes :
            index_count = self._get(txn, METADATA_VERTEX_INDEX_COUNTER)
            for index_id in range(0, index_count):
                fields, filter_tuples = self._get(txn, build_key(METADATA_VERTEX_INDEX_FIELDS_PREFIX, index_id))
                filters = dict(filter_tuples)
                if fields != UNDEFINED:
                    prefix = build_key(KIND_INDEX, KIND_VERTEX, index_id)
                    self._node_indexes.append(ExactIndex(self, KIND_VERTEX, prefix, *fields, **filters))
            #   2nd edge indexes (~ same code, but I'm not sure it makes sense to factor...)
            index_count = self._get(txn, METADATA_EDGE_INDEX_COUNTER)
            for index_id in range(0, index_count):
                fields, filter_tuples = self._get(txn, build_key(METADATA_EDGE_INDEX_FIELDS_PREFIX, index_id))
                filters = dict(filter_tuples)
                if fields != UNDEFINED:
                    prefix = build_key(KIND_INDEX, KIND_EDGE, index_id)
                    self._edge_indexes.append(ExactIndex(self, KIND_EDGE, prefix, *fields, **filters))
            self._transaction_commit(txn)
        except Exception as e:  # pragma : no cover
            self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))
        # Instantiate an optimizer (which will choose between scan and sequential scan AND (in the future) cache best index for most runned queries)
        self._optimizer = Optimizer(self)

    def _prepare_database(self, txn):
        self._set(txn, PREPARED, 1)
        self._set(txn, METADATA_VERTEX_COUNTER, 0)
        self._set(txn, METADATA_VERTEX_INDEX_COUNTER, 0)
        self._set(txn, METADATA_VERTEX_REMOVED_COUNTER, 0)
        self._set(txn, METADATA_EDGE_COUNTER, 0)
        self._set(txn, METADATA_EDGE_REMOVED_COUNTER, 0)
        self._set(txn, METADATA_EDGE_INDEX_COUNTER, 0)

    def _ensure_prepared(self):
        if not(self._has_key(PREPARED)):
            txn = self._transaction_begin()
            try:
                self._prepare_database(txn)
                self._transaction_commit(txn)
            except Exception as e:
                self._transaction_rollback(txn)
                raise GrapheekDataPreparationFailedException(repr(e))
        self.__initialize__()

    # Following methods MUST be overriden by child classes :

    def _db_close(self):
        raise NotImplementedError

    def _transaction_begin(self):
        raise NotImplementedError

    def _transaction_commit(self, handle):
        raise NotImplementedError

    def _transaction_rollback(self, handle):
        raise NotImplementedError

    def _has_key(self, key):
        raise NotImplementedError

    def _get(self, txn, key):
        raise NotImplementedError

    def _bulk_get(self, txn, keys):
        raise NotImplementedError

    def _set(self, key, value):
        raise NotImplementedError

    def _bulk_set(self, key, value):
        raise NotImplementedError

    def _remove(self, *key):
        raise NotImplementedError

    def _bulk_remove(self, txn, keys):
        raise NotImplementedError

    def _remove_prefix(self, prefix):
        raise NotImplementedError

    # For next methods, default implementation MAY suffice :
    # BUT overriding them in child classes is a path to performance :)
    # See for instance, kyotocab.KyotoCabinetGraph _init_lst, _get_lst & _append_to_lst overriding

    def _init_lst(self, txn, key):
        # Create an empty list
        self._set(txn, key, [])

    def _get_lst(self, txn, key):
        res = self._get(txn, key)  # if _get returns UNDEFINED, that's also ok, I return it...
        return res

    def _set_lst(self, txn, key, values):
        self._set(txn, key, values)

    def _bulk_get_lst(self, txn, keys):
        return [self._get(txn, key) for key in keys]

    def _append_to_lst(self, txn, key, value):
        lst = self._get(txn, key)
        if lst == UNDEFINED:
            lst = []
        self._set(txn, key, lst + [value])

    def _bulk_append_to_lst(self, txn, key, values):
        lst = self._get(txn, key)
        if lst == UNDEFINED:
            lst = []
        self._set(txn, key, lst + values)

    def _remove_from_lst(self, txn, key, value):
        old = self._get(txn, key)
        # Caution : we are only removing ONE occurence
        # This is voluntary
        # For instance, it lst contains neighbour node, we need to remove only one occurence
        # cause current entity and neighbour node can be linked multiple time
        new = old[:]
        new.remove(value)
        self._set(txn, key, new)

    def _bulk_remove_from_lst(self, txn, key, values):
        old = self._get(txn, key)
        new = old[:]
        for value in values:
            new.remove(value)
        self._set(txn, key, new)

    def _update_inc(self, txn, key, value=1):
        oldval_ = self._get(txn, key)
        if oldval_ == UNDEFINED:
            raise GrapheekMissingKeyException
        oldval = int(oldval_)
        newval = oldval + value
        self._set(txn, key, newval)
        return newval

    def _update_dec(self, txn, key, value=1):
        return self._update_inc(txn, key, value * -1)

    def _new_id_for_key(self, txn, key):
        return self._update_inc(txn, key) - 1  # -1 so that entity idx starts at 0

    # Sub data update (no need to override, IMHO)
    # Remark that we are using kind, id instead of key (key is a *key* of the Key Value Store, id is an entity id (node id or edge id for instance))

    def _get_data(self, txn, kind, entity_id, *args, **kwargs):
        key = build_key(kind, entity_id, DATA_SUFFIX)
        result = self._get(txn, key)  # if _get returns UNDEFINED, that's also ok, I return it...
        if args:
            tmpres = dict((k, v) for k, v in result.items() if k in args)
            result = tmpres
            if kwargs:
                flat = kwargs.get('flat', False)
                if flat:
                    tmpres = [result[k] for k in args]
                    result = tmpres
        return result

    def _bulk_update_data(self, _txn, _kind, _entity_id, **updates):
        check_valid_data(updates)
        release_txn = False
        if _txn is None:
            _txn = self._transaction_begin()
            release_txn = True
        try:
            data = self._get_data(_txn, _kind, _entity_id)
            # Remove entity from indexes before updating data (we'll re-add it just after with new values)
            self._remove_from_all_entity_indexes(_txn, _kind, _entity_id)
            # Do the update :
            data.update(**updates)
            key = build_key(_kind, _entity_id, DATA_SUFFIX)
            self._set(_txn, key, data)
            # Readding entity to indexes :
            self._add_to_all_entity_indexes(_txn, _kind, _entity_id, data)
            if release_txn:
                self._transaction_commit(_txn)
        except Exception as e:
            if release_txn:
                self._transaction_rollback(_txn)
            raise GrapheekDataException(repr(e))

    def _update_data(self, kind, entity_id, subkey, value):
        self._bulk_update_data(None, kind, entity_id, **{subkey: value})  # Don't create a txn, it will be created by _bulk_update_data

    # DON'T OVERRIDE THOSE METHODS :

    def _get_server_script(self, _fname):
        # 1st, get function from cache :
        func = self._script_cache.get(_fname, None)
        if func is None:
            for module in self._script_module_cache:
                try:
                    func = getattr(module, _fname)
                except AttributeError:
                    pass  # it may not exist in module but may exist in exist...
                if func is not None:
                    # Updating cache and stop iteration
                    self._script_cache[_fname] = func
                    break
        if func is None:
            # not found -> raise an exception to warn user
            raise GrapheekUnknownScriptException('%s method cannot be found' % (_fname,))
        return func

    def _add_node(self, txn, data, node_id=None):
        assert(isinstance(data, dict))
        release_txn = False
        if txn is None:
            release_txn = True
            txn = self._transaction_begin()
        try:
            operation = Addition()
            if node_id is None:
                node_id = self._new_id_for_key(txn, METADATA_VERTEX_COUNTER)
            node = Node(node_id, self)
            operation.set(build_key(KIND_VERTEX, node_id, DATA_SUFFIX), data.copy())
            operation.merge(node._add_denorm_data())
            # Adding node id to proper node ids list :
            operation.append_to_lst(build_key(METADATA_VERTEX_ID_LIST_PREFIX, node_id // CHUNK_SIZE), node_id)
            if release_txn:
                # Applying operation (before updating indexes as indexes needs key to be updated)
                operation.apply(txn, self)
                # Updating vertex indexes :
                self._add_to_all_entity_indexes(txn, KIND_VERTEX, node_id, data)
                # Updating denorm node count
                self._node_count += 1
                self._transaction_commit(txn)
                return node
            else:
                return operation
        except Exception as e:
            if release_txn:
                self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def _bulk_add_node(self, node_defns, silent=False):
        txn = self._transaction_begin()
        try:
            operation = Addition()
            node_count = len(node_defns)
            last_node_id = self._update_inc(txn, METADATA_VERTEX_COUNTER, node_count) - 1
            first_node_id = last_node_id - node_count + 1
            range_ids = list(range(first_node_id, last_node_id + 1))
            for node_id, data in zip(range_ids, node_defns):
                operation.merge(self._add_node(txn, data, node_id))
            operation.apply(txn, self)
            # now updating indexes :
            self._bulk_add_to_all_entity_indexes(txn, KIND_VERTEX, range_ids)
            # update denormalized node count
            self._node_count += node_count
            # all is ok, commit :)
            self._transaction_commit(txn)
            if not(silent):
                res = [Node(node_id, self) for node_id in range_ids]
                return res
        except Exception as e:
            self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def _remove_node(self, node_id):
        txn = self._transaction_begin()
        try:
            operation = Removal()
            node = Node(node_id, self)
            # edges attached to this node must be removed before removing any node data
            # (this will infer some denorm data modifications on related outer and inner nodes)
            out_edge_ids = self._get_lst(txn, build_key(KIND_VERTEX, node_id, OUT_EDGES_SUFFIX))
            for out_edge_id in out_edge_ids:
                operation.merge(self._remove_edge(out_edge_id, txn=txn))
            in_edge_ids = self._get_lst(txn, build_key(KIND_VERTEX, node_id, IN_EDGES_SUFFIX))
            for in_edge_id in in_edge_ids[:]:
                operation.merge(self._remove_edge(in_edge_id, txn=txn))
            # now, removing node related data and denorm data :
            operation.merge(node._remove_denorm_data())
            # Removing node id from proper node ids list :
            operation.remove_from_lst(build_key(METADATA_VERTEX_ID_LIST_PREFIX, int(node_id) // CHUNK_SIZE), node_id)
            # Updating vertex indexes before removing data
            self._remove_from_all_entity_indexes(txn, KIND_VERTEX, node_id)
            # Now removing data
            operation.remove(build_key(KIND_VERTEX, node_id, DATA_SUFFIX))
            # Incrementing removed vertex counter :
            operation.update_inc(METADATA_VERTEX_REMOVED_COUNTER)
            # Applying operation
            operation.apply(txn, self)
            # Update denorm counter :
            self._node_count -= 1
            self._transaction_commit(txn)
        except Exception as e:
            self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def _add_edge(self, txn, source, target, data=None, edge_id=None):
        assert(isinstance(data, dict))
        release_txn = False
        if txn is None:
            release_txn = True
            txn = self._transaction_begin()
        try:
            operation = Addition()
            if edge_id is None:
                edge_id = self._new_id_for_key(txn, METADATA_EDGE_COUNTER)
            data = data or {}
            # Initializing
            operation.init_lst(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX))
            operation.init_lst(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX))
            operation.init_lst(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX))
            # Persistence and denormalization - START
            operation.set(build_key(KIND_EDGE, edge_id, DATA_SUFFIX), data)
            # Adding edge id to proper edge ids list :
            operation.append_to_lst(build_key(METADATA_EDGE_ID_LIST_PREFIX, int(edge_id) // CHUNK_SIZE), edge_id)
            # Edge data denormalization
            edge = Edge(edge_id, self)
            operation.merge(edge._add_denorm_data(source, target))
            # source data denormalization
            operation.merge(source._add_denorm_src_data(target, edge))
            # target data denormalization
            operation.merge(target._add_denorm_tgt_data(source, edge))
            if release_txn:
                # Applying operation (before updating indexes as indexes needs key to be updated)
                operation.apply(txn, self)
                # Updating edge indexes :
                self._add_to_all_entity_indexes(txn, KIND_EDGE, edge_id, data)
                # Updating edge count denorm :
                self._edge_count += 1
                self._transaction_commit(txn)
                # Persistence and denormalization - END
                return edge
            else:
                return operation
        except Exception as e:
            if release_txn:
                self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def _bulk_add_edge(self, edge_defns, silent=False):
        txn = self._transaction_begin()
        try:
            operation = Addition()
            edge_count = len(edge_defns)
            last_edge_id = self._update_inc(txn, METADATA_EDGE_COUNTER, edge_count) - 1
            first_edge_id = last_edge_id - edge_count + 1
            range_ids = list(range(first_edge_id, last_edge_id + 1))
            for edge_id, edge_defn in zip(range_ids, edge_defns):
                source, target, data = edge_defn
                operation.merge(self._add_edge(txn, source, target, data, edge_id=edge_id))
            operation.apply(txn, self)
            # now updating indexes :
            self._bulk_add_to_all_entity_indexes(txn, KIND_EDGE, range_ids)
            # update denormalized node count
            self._edge_count += edge_count
            # all is ok, commit :)
            self._transaction_commit(txn)
            if not(silent):
                return [Edge(edge_id, self) for edge_id in range_ids]
        except Exception as e:
            self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    def _bulk_add_edge_by_ids(self, edge_id_defns, silent=False):
        for _, _, data in edge_id_defns:
            check_valid_data(data)
        edge_defns = [(self.v(source_id), self.v(target_id), data) for source_id, target_id, data in edge_id_defns]
        return self._bulk_add_edge(edge_defns, silent)

    def _add_edge_by_ids(self, source_id, target_id, data=None):
        source = Node(source_id, self)
        target = Node(target_id, self)
        return self._add_edge(None, source, target, data)

    def _remove_edge(self, edge_id, txn=None):
        def remove_vertex_denorm_data(node, other, node_suffix):
            rm_denorm_method = node._remove_denorm_src_data if node_suffix == OUT_VERTICES_SUFFIX else node._remove_denorm_tgt_data
            return rm_denorm_method(other, edge)

        release_txn = False
        if txn is None:
            release_txn = True
            txn = self._transaction_begin()
        try:
            operation = Removal()
            edge = Edge(edge_id, self)
            source = target = None
            source_id = self._get_lst(txn, build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX))[0]
            source = Node(source_id, self)
            target_id = self._get_lst(txn, build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX))[0]
            target = Node(target_id, self)
            operation.merge(remove_vertex_denorm_data(source, target, OUT_VERTICES_SUFFIX))
            operation.merge(remove_vertex_denorm_data(target, source, IN_VERTICES_SUFFIX))
            # Removing edge id from proper edge ids list :
            operation.remove_from_lst(build_key(METADATA_EDGE_ID_LIST_PREFIX, edge_id // CHUNK_SIZE), edge_id)
            # Removing edge related lists
            operation.remove(build_key(KIND_EDGE, edge_id, IN_VERTICES_SUFFIX))
            operation.remove(build_key(KIND_EDGE, edge_id, OUT_VERTICES_SUFFIX))
            operation.remove(build_key(KIND_EDGE, edge_id, BOTH_VERTICES_SUFFIX))
            # Updating edge indexes before removing data
            self._remove_from_all_entity_indexes(txn, KIND_EDGE, edge_id)
            # Now removing edge data
            operation.remove(build_key(KIND_EDGE, edge_id, DATA_SUFFIX))
            # Removing edge denormalized data
            assert((source or target) is not None)  # Both source and target must NOT be None
            operation.merge(edge._remove_denorm_data(source, target))
            # Incrementing removed edge counter :
            operation.update_inc(METADATA_EDGE_REMOVED_COUNTER)
            # Updating edge counter denorm :
            self._edge_count -= 1
            if release_txn:
                operation.apply(txn, self)
                self._transaction_commit(txn)
            else:
                return operation
        except Exception as e:
            if release_txn:
                self._transaction_rollback(txn)
            raise GrapheekDataException(repr(e))

    # Item id -> Item helper

    def _item_from_id(self, kind, entity_id):
        Klass = Node if kind == KIND_VERTEX else Edge
        return Klass(entity_id, self)

    # Traversals given entity ids :

    def _neighbors(self, _kind, _entity_id, _traversal, _cache=None, _random=False, **filters):
        """
        _neighbors is using a cache (transmitted by iterator) to speed up lookup
        This is done to asymptotically improve performance at the expense of local memory
        Why asymptotically ? because for short path traversal, cache has few probability to be hitted
        On the contrary, for "long path", probability is greater

        WARNING : filter is only used to check if index can be used
        It is not used to really filter neighbors (which is the EntityIterator responsibily)
        So don't be confused by this function signature :)
        """
        # seq scan on neighbours :
        #   potentially using cache
        #   building neighbour_ids
        has_cache = _cache is not None
        neighbour_ids = None
        if has_cache:
            cache_key = (_kind, _entity_id, _traversal)
            neighbour_ids = _cache.get(cache_key, None)
        if neighbour_ids is None:
            # no cache..
            key = build_key(_kind, _entity_id, _traversal)
            neighbour_ids = self._get_lst(None, key)
            if has_cache:
                _cache[cache_key] = neighbour_ids
        # Special case when filters is empty or kind is edge (no need to look in index in the case of edge)
        if (_kind == KIND_EDGE) or not filters:
            if not(_random):
                for entity_id in iter(neighbour_ids):
                    yield entity_id
            else:
                yield random.choice(neighbour_ids)
        else:
            # Preparing metrics that will allow to choose between sequential access and index access :
            iterator = None
            index_count = len(self._node_indexes) if _kind == KIND_VERTEX else len(self._edge_indexes)
            seq_count = self._get(None, build_key(_kind, _entity_id, _traversal, COUNT_SUFFIX))
            # querying index needs <index_count> key access
            # challenging seq_count
            if seq_count > 2 * index_count:
                # ---
                # Why "2" in next line ?
                # Because if an index use appears to be better than index scan, we will have
                # to challenge every indexed ids to check that it is among current entity neighbours
                # so it will lead to
                # which can cost a lot (<index_count> disk access for persistent backends)
                _traversal_kind = {
                    IN_VERTICES_SUFFIX: KIND_VERTEX,
                    OUT_VERTICES_SUFFIX: KIND_VERTEX,
                    BOTH_VERTICES_SUFFIX: KIND_VERTEX,
                    IN_EDGES_SUFFIX: KIND_EDGE,
                    OUT_EDGES_SUFFIX: KIND_EDGE,
                    BOTH_EDGES_SUFFIX: KIND_EDGE,
                }[_traversal]
                iterator = self._optimizer.index_or_seq_scan_iterator(_traversal_kind, seq_count * 2, **filters)
            if iterator is None:
                if not(_random):
                    for entity_id in iter(neighbour_ids):
                        yield entity_id
                else:
                    yield random.choice(neighbour_ids)
            else:
                indexed_ids = set(iterator)
                if not(_random):
                    for entity_id in indexed_ids.intersection(neighbour_ids):
                        yield entity_id
                else:  # pragma : no cover
                    yield random.choice(indexed_ids.intersection(neighbour_ids))

    # Don't override next methods : it's the public interface
    # (or some "private" helper methods for public methods)

    def __X(self, _kind, **filters):

        if self._closed:
            raise GrapheekDataException('Graph is closed')
        if _kind == KIND_VERTEX:
            seq_count = self._node_count
        else:
            seq_count = self._edge_count
        iterator = self._optimizer.index_or_seq_scan_iterator(_kind, seq_count, **filters)
        if iterator is None:  # no index is better than sequential scan, so choose it
            iterator = self._optimizer.get_kind_ids(None, _kind)

        return EntityIterator(self, _kind, filters, iterator)

    def V(self, **filters):
        return self.__X(KIND_VERTEX, **filters)

    def E(self, **filters):
        return self.__X(KIND_EDGE, **filters)

    def v(self, id):
        if self._closed:  # pragma : no cover
            raise GrapheekDataException('Graph is closed')
        return Node(id, self)

    def e(self, id):
        if self._closed:  # pragma : no cover
            raise GrapheekDataException('Graph is closed')
        return Edge(id, self)

    def add_node(self, **kwargs):
        check_valid_data(kwargs)
        return self._add_node(None, kwargs.copy())

    def bulk_add_node(self, node_defns, silent=False):
        for data in node_defns:
            check_valid_data(data)
        node_defns_copy = [node_defn.copy() for node_defn in node_defns]
        return self._bulk_add_node(node_defns_copy, silent)

    def add_edge(self, _source, _target, **kwargs):
        check_base_node(_source)
        check_base_node(_target)
        check_valid_data(kwargs)
        return self._add_edge(None, _source, _target, kwargs.copy())

    def bulk_add_edge(self, edge_defns, silent=False):
        try:
            for source, target, data in edge_defns:
                check_base_node(source)
                check_base_node(target)
                check_valid_data(data)
        except (TypeError, ValueError) as e:
            raise GrapheekDataException(repr(e))
        edge_defns_copy = [(source, target, data.copy()) for source, target, data in edge_defns]
        return self._bulk_add_edge(edge_defns_copy, silent)

    def bulk_add_edge_by_ids(self, edge_id_defns, silent=False):
        return self._bulk_add_edge_by_ids(edge_id_defns, silent)

    def add_edge_by_ids(self, _source_id, _target_id, **kwargs):
        return self._add_edge_by_ids(_source_id, _target_id, kwargs)

    def update_data(self, kind, entity_id, attr, value):
        self._update_data(kind, entity_id, attr, value)

    def _add_to_all_entity_indexes(self, txn, kind, entity_id, data):
        indexes = self._node_indexes if kind == KIND_VERTEX else self._edge_indexes
        for index in indexes:
            index.add(txn, entity_id, data)

    def _bulk_add_to_all_entity_indexes(self, txn, kind, entity_ids):
        indexes = self._node_indexes if kind == KIND_VERTEX else self._edge_indexes
        for index in indexes:
            index.bulk_add(txn, iter(entity_ids))

    def _remove_from_all_entity_indexes(self, txn, kind, entity_id):
        indexes = self._node_indexes if kind == KIND_VERTEX else self._edge_indexes
        for index in indexes:
            index.remove(txn, entity_id)

    def _add_entity_index(self, _kind, *args, **filters):
        fields = list(args)
        fields.sort()
        filter_tuples = list(filters.items())
        filter_tuples.sort()
        index_signature = [fields, filter_tuples]
        index_signature_string = normalize_value(index_signature)
        assert(_kind in (KIND_VERTEX, KIND_EDGE))
        if _kind == KIND_VERTEX:
            METADATA_INDEX_PREFIX = METADATA_VERTEX_INDEX_PREFIX
            METADATA_INDEX_COUNTER = METADATA_VERTEX_INDEX_COUNTER
            METADATA_COUNTER = METADATA_VERTEX_COUNTER
            METADATA_INDEX_LIST = METADATA_VERTEX_INDEX_LIST
            METADATA_INDEX_FIELDS_PREFIX = METADATA_VERTEX_INDEX_FIELDS_PREFIX
            entity_indexes = self._node_indexes
        else:
            METADATA_INDEX_PREFIX = METADATA_EDGE_INDEX_PREFIX
            METADATA_INDEX_COUNTER = METADATA_EDGE_INDEX_COUNTER
            METADATA_COUNTER = METADATA_EDGE_COUNTER
            METADATA_INDEX_LIST = METADATA_EDGE_INDEX_LIST
            METADATA_INDEX_FIELDS_PREFIX = METADATA_EDGE_INDEX_FIELDS_PREFIX
            entity_indexes = self._edge_indexes
        # Checking if index already exists (won't create if it is the case) :
        index_key = '/'.join([METADATA_INDEX_PREFIX] + [index_signature_string])
        index_id = self._get(None, index_key)
        if index_id != UNDEFINED:
            # Index already exists
            raise GrapheekIndexAlreadyExistsException
        txn = self._transaction_begin()
        try:
            index_id = self._new_id_for_key(txn, METADATA_INDEX_COUNTER)
            self._set(txn, index_key, index_id)
            self._append_to_lst(txn, METADATA_INDEX_LIST, index_id)  # "Registering" index in index list
            self._set(txn, build_key(METADATA_INDEX_FIELDS_PREFIX, index_id), index_signature)
            entity_count = self._get(txn, METADATA_COUNTER)
            if entity_count == UNDEFINED:
                raise GrapheekDataException
            prefix = build_key(KIND_INDEX, _kind, index_id)
            # NOTE : So far, only ExactIndex is supported :
            index = ExactIndex(self, _kind, prefix, *fields, **filters)
            # Entity iterator : index will iterate on it to create denormalized data
            # that will (should ;)) speedup query
            id_iterator = self._optimizer.get_kind_ids(txn, _kind)
            # Building index :
            index.bulk_add(txn, id_iterator)
            # Everything seems to be ok, adding index to proper list
            entity_indexes.append(index)
            self._transaction_commit(txn)
        except Exception as e:
            self._transaction_rollback(txn)
            raise GrapheekIndexCreationFailedException(repr(e))

    def add_node_index(self, *fields, **filters):
        self._add_entity_index(KIND_VERTEX, *fields, **filters)

    def add_edge_index(self, *fields, **filters):
        self._add_entity_index(KIND_EDGE, *fields, **filters)

    def get_node_indexes(self):
        return [[index._fields, dict(index._filter_items)] for index in self._node_indexes]

    def get_edge_indexes(self):
        return [[index._fields, dict(index._filter_items)] for index in self._edge_indexes]

    def _remove_entity_index(self, _kind, *args, **kwargs):
        assert(_kind in (KIND_VERTEX, KIND_EDGE))
        fields = list(args)
        fields.sort()
        filter_tuples = list(kwargs.items())
        filter_tuples.sort()
        index_signature = [fields, filter_tuples]
        index_signature_string = normalize_value(index_signature)
        if _kind == KIND_VERTEX:
            indexes = self._node_indexes
            METADATA_INDEX_PREFIX = METADATA_VERTEX_INDEX_PREFIX
            METADATA_INDEX_LIST = METADATA_VERTEX_INDEX_LIST
            METADATA_INDEX_FIELDS_PREFIX = METADATA_VERTEX_INDEX_FIELDS_PREFIX
            METADATA_INDEX_COUNTER = METADATA_VERTEX_INDEX_COUNTER
        else:
            indexes = self._edge_indexes
            METADATA_INDEX_PREFIX = METADATA_EDGE_INDEX_PREFIX
            METADATA_INDEX_LIST = METADATA_EDGE_INDEX_LIST
            METADATA_INDEX_FIELDS_PREFIX = METADATA_EDGE_INDEX_FIELDS_PREFIX
            METADATA_INDEX_COUNTER = METADATA_EDGE_INDEX_COUNTER
        index_to_remove = None
        for index in indexes:
            if index._fields == fields and index._filter_items == set(kwargs.items()):
                index_to_remove = index
                break
        if index_to_remove is None:
            raise GrapheekIndexRemovalFailedException
        txn = self._transaction_begin()
        try:
            # Removing index
            index.delete(txn)  # it is the responsability of the index to remove everything it created
            # Removing index from index list :
            indexes.remove(index)
            # update counter
            self._update_dec(txn, METADATA_INDEX_COUNTER)
            # Removing "management key"
            index_key = '/'.join([METADATA_INDEX_PREFIX] + [index_signature_string])
            index_id = self._get(txn, index_key)
            # Remove index key :
            self._remove(txn, index_key)
            # Removing index fields :
            self._remove(txn, build_key(METADATA_INDEX_FIELDS_PREFIX, index_id))
            # Removing index from index list :
            self._remove_from_lst(txn, METADATA_INDEX_LIST, index_id)
            self._transaction_commit(txn)
        except Exception as e:
            self._transaction_rollback(txn)
            raise GrapheekIndexRemovalFailedException(repr(e))

    def remove_node_index(self, *fields, **filters):
        self._remove_entity_index(KIND_VERTEX, *fields, **filters)

    def remove_edge_index(self, *fields, **filters):
        self._remove_entity_index(KIND_EDGE, *fields, **filters)

    def _operator_kind_and_checks(self, *entity_iterators):
        # ensure entity iterators are of the same kind :
        kinds = set([it._src_kind for it in entity_iterators])
        if len(set(kinds)) > 1:
            raise GrapheekMixedKindException
        kind = kinds.pop()
        return kind

    def union(self, *entity_iterators):
        kind = self._operator_kind_and_checks(*entity_iterators)

        def id_iterator(*entity_iterators):
            known_ids = set()
            clone_entity_iterators = [entity_iterator._clone_initial() for entity_iterator in entity_iterators]
            for entity in chain(*clone_entity_iterators):
                entity_id = entity._entity_id
                if entity_id not in known_ids:
                    known_ids.add(entity_id)
                    yield entity_id
        return EntityIterator(self, kind, {}, id_iterator(*entity_iterators))

    def _marshalled_operator_helper(self, commands):  # pragma : no cover
        entity_iterators = []
        for command in commands:
            method_name, args, kwargs = command[0]
            method = getattr(self, method_name)
            entity_iterators.append(method(*args, **kwargs))
        return entity_iterators

    def _marshalled_union(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.union(*entity_iterators)

    def _set_operation(self, func, *entity_iterators):
        kind = self._operator_kind_and_checks(*entity_iterators)
        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
        ids = reduce(func, id_sets)
        return EntityIterator(self, kind, {}, iter(ids))

    def issubset(self, *entity_iterators):
        self._operator_kind_and_checks(*entity_iterators)
        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
        func = lambda x, y: x.issubset(y)
        value = reduce(func, id_sets)
        return value

    def _marshalled_issubset(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.issubset(*entity_iterators)

    def issuperset(self, *entity_iterators):
        self._operator_kind_and_checks(*entity_iterators)
        id_sets = [set(it._clone_initial().ids()) for it in entity_iterators]
        func = lambda x, y: x.issuperset(y)
        value = reduce(func, id_sets)
        return value

    def _marshalled_issuperset(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.issuperset(*entity_iterators)

    def intersection(self, *entity_iterators):
        func = lambda x, y: x.intersection(y)
        return self._set_operation(func, *entity_iterators)

    def _marshalled_intersection(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.intersection(*entity_iterators)

    def difference(self, *entity_iterators):
        func = lambda x, y: x.difference(y)
        return self._set_operation(func, *entity_iterators)

    def _marshalled_difference(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.difference(*entity_iterators)

    def symmetric_difference(self, *entity_iterators):
        func = lambda x, y: x.symmetric_difference(y)
        return self._set_operation(func, *entity_iterators)

    def _marshalled_symmetric_difference(self, *commands, **kwargs):  # pragma : no cover
        entity_iterators = self._marshalled_operator_helper(commands)
        return self.symmetric_difference(*entity_iterators)

    def setup_server_scripts(self, *module_paths):
        import sys
        # Following lines may raise an exception if data module cannot be imported
        # I'm letting exception propagate so that the user can fix path (or other potential errors)
        self._script_module_cache = []
        for module_path in list(module_paths) + ['grapheekdb.server.scripts']:
            if module_path:
                __import__(module_path)
                class_module = sys.modules[module_path]
                self._script_module_cache.append(class_module)
        self._script_cache = {}



# -*- coding:utf-8 -*-

from setuptools import setup, find_packages
from grapheekdb import __version__

setup(
    name="grapheekdb",
    version=__version__,
    author="Raphaël Braud",
    author_email="grapheekdb@gmail.com",
    packages=find_packages(),
    zip_safe=False,
    url="https://bitbucket.org/nidusfr/grapheekdb",
    license="GPL v3",
    description="GrapheekDB is a pure Python graph database which is fast and lightweight",
    long_description=open("README.rst").read(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
    ],
    install_requires=[
        "pyzmq",
        "msgpack-python==0.4.0"
    ],
    entry_points="""
    [console_scripts]
    grapheekserve = grapheekdb.server.serve:main
    """
)



#!/usr/bin/env
# -*- coding: utf-8 -*-

from collections import defaultdict


class Addition(object):

    def __init__(self):
        self._append_to_lst_registry = defaultdict(list)
        self._update_inc_registry = defaultdict(int)
        self._init_lst_registry = set()
        self._set_registry = {}
        self._applied = False

    def merge(self, other):
        for key, entity_ids in other._append_to_lst_registry.items():
            self._append_to_lst_registry[key].extend(entity_ids)
        for key, value in other._update_inc_registry.items():
            self._update_inc_registry[key] += value
        self._init_lst_registry.update(other._init_lst_registry)
        self._set_registry.update(other._set_registry)

    def append_to_lst(self, key, entity_id):
        self._append_to_lst_registry[key].append(entity_id)

    def update_inc(self, key):
        self._update_inc_registry[key] += 1

    def init_lst(self, key):
        self._init_lst_registry.add(key)

    def set(self, key, value):
        self._set_registry[key] = value

    def apply(self, txn, graph):
        assert(not(self._applied))
        # Init lst :
        for key in self._init_lst_registry:
            graph._init_lst(txn, key)
        # Add entity_ids to lists
        for key, entity_ids in self._append_to_lst_registry.items():
            graph._bulk_append_to_lst(txn, key, entity_ids)
        # Adding key value :
        graph._bulk_set(txn, self._set_registry)
        # ---
        self._applied = True
        # Increasing value
        for key, value in self._update_inc_registry.items():
            graph._update_inc(txn, key, value)


class Removal(object):

    def __init__(self):
        self._remove_from_lst_registry = defaultdict(list)
        self._update_inc_registry = defaultdict(int)
        self._update_dec_registry = defaultdict(int)
        self._remove_registry = set()
        self._applied = False

    def merge(self, other):
        for key, entity_ids in other._remove_from_lst_registry.items():
            self._remove_from_lst_registry[key].extend(entity_ids)
        for key, value in other._update_inc_registry.items():
            self._update_inc_registry[key] += value
        for key, value in other._update_dec_registry.items():
            self._update_dec_registry[key] += value
        self._remove_registry.update(other._remove_registry)

    def remove_from_lst(self, key, entity_id):
        self._remove_from_lst_registry[key].append(entity_id)

    def update_inc(self, key):
        self._update_inc_registry[key] += 1

    def update_dec(self, key):
        self._update_dec_registry[key] += 1

    def remove(self, key):
        self._remove_registry.add(key)

    def apply(self, txn, graph):
        assert(not(self._applied))
        # Removing entity_ids from lists
        for key, entity_ids in self._remove_from_lst_registry.items():
            graph._bulk_remove_from_lst(txn, key, entity_ids)
        # Increasing value
        for key, value in self._update_inc_registry.items():
            graph._update_inc(txn, key, value)
        # Decreasing value
        for key, value in self._update_dec_registry.items():
            graph._update_dec(txn, key, value)
        # Removing keys :
        graph._bulk_remove(txn, self._remove_registry)
        # ---
        self._applied = True
